#######
#This file contains a collection of classes and functions which support specReduce.py
#######


###  debug imports
from pprint import pprint

import math
import numpy as np
import warnings
from astropy.io import fits
#import matplotlib
import astropy
import cosmics
import matplotlib.pyplot as plt
import copy
import os
import scipy
import scipy.optimize
import scipy.ndimage
import json



import USERDefs  #sets up parameters for this instrument and use case


testing_display=True

class Frame:
  """basic container for SpecReduce data frames.

  tracks orientation of the data block, header cards, and provides several functions for handling frames
  list of loadmodes: default,raw,badpix,badpixDebias,badpixDebiasDecosmic
  list of types: lamp,data,bias,flat,response
  """


  #Frame Class todo list
  #  copy constructor
  #  constructor from bodyarray

  def __init__(self,Filename=None,loadMode='Default',type=None,frameSetup=None,bias=None,savePath=USERDefs.savepath):
    """
    constructor for a frame.

    normally will be used for loading a frame from a file
    """
    if (frameSetup is None):frameSetup = {}
    self.originalFile=None
    self.type=type
    self.roll=0
    self.pitch=0
    self.yaw=0
    #self.rawFits=None
    if Filename:
      self.load(Filename,loadMode=loadMode,type=self.type,frameSetup=frameSetup,bias=bias,savePath=savePath)
    else:
      self.originalFile = "Frame_created_without_read"
      self.origpath=None
      self.frameSetup= frameSetup

#  def __deepcopy__(self,memo):
#    print "DEBUG: deepcopying a frame using internal method"
#    cop=Frame()
#    print "built a frame"
#    cop.frameSetup=copy.deepcopy(self.frameSetup,memo)
#    print "have the setup"
#    cop.originalFile=self.originalFile
#    print "got the orig"
#    cop.type=self.type
#    print "type done"
#    cop.roll=self.roll
#    cop.pitch=self.pitch
#    cop.yaw=self.yaw
#    print "about to do the raw"
#    cop.rawFits=fits.open(cop.originalFile)
#    #cop.rawFits=copy.deepcopy(self.rawFits,memo)
#    print "did the rawFits"
#    try:
#      print "deepcopy data"
#      cop.Data = copy.deepcopy(self.Data,memo)
#    except:
#      pass
#    try:
#      print "deepcopy head"
#      cop.Head = copy.deepcopy(self.Head,memo)
#    except:
#      pass
#    print "returning from deepcopy"
#    return cop

  def load(self,originalFile=None,loadMode='Default',type=None,axis=0,frameSetup=None,bias=None,savePath=USERDefs.savepath):
    """
    load a fits frame from a file.

    using the frameSetup keyword argument you can control the orientation of the frame and track some parameters of the instrument
    """
    print "working in mode ",loadMode, " with frameSetup ",frameSetup
    if (frameSetup is None):frameSetup = {}
    self.originalFile=originalFile
    #self.rawFits = fits.open(originalFile)
    rawFits = fits.open(originalFile)
    self.frameSetup= frameSetup
    #self.Data = np.array(self.rawFits[0].data)
    #self.Head = fits.Header(self.rawFits[0].header)
    self.Data = np.array(rawFits[0].data)
    self.Head = fits.Header(rawFits[0].header)

    #deal with rotation and reflection
    if ('yaw' in frameSetup and frameSetup['yaw']):
      #print "DEBUG: rotating input"
      self.switchAxis()
    if ('pitch' in frameSetup and frameSetup['pitch']):
      #print "DEBUG: flipping input"
      self.switchDirection()
    if ('roll' in frameSetup and frameSetup['roll']):
      #print "DEBUG: flipping input"
      self.switchDirection()

    if (loadMode.lower() == 'default' or  'badpix' in loadMode.lower() ):
      self.Data=badPix(self.Data,BPvalue=USERDefs.BadPixelValue)
      self.Head['history']="bad pixels removed."

    if (bias != "*none" and ('debias' in loadMode.lower() or bias == "*over")):
      if (bias==None or bias=="*over"):
        print "generate bias from overscan"
        bias = self.makeBiasFromOver(self.frameSetup['overscan'],self.frameSetup['biasaxis'])
        if (USERDefs.saveall):
          path,fn,ext = splitPathExt(self.originalFile)
          #outfn = savePath+fn+"_bias"+ext
          outfn = os.path.join(savePath,fn+"_bias"+ext)
          bias.toFile(outfn)
      self.Data = self.Data-bias.Data
      self.Head['history'] = "debiased"
      if (USERDefs.saveall):
        path,fn,ext = splitPathExt(self.originalFile)
        #outfn = savePath+fn+"_debiased"+ext
        outfn = os.path.join(savePath,fn+"_debiased"+ext)
        self.toFile(outfn)

     #TODO  - decosmic on load...



  #should check for list of standard header cards to 
  def switchAxis(self): #yaw
    """rotates a frame 90 degrees

    this allows the user to process all frames in a similar fasion by providing a way to orient all frames the same.
    """
    print "switching axis"
    self.yaw+=1
    if (self.yaw>3):
      self.yaw-=4
    self.Data = np.rot90(self.Data)
    axissize1 = self.Head['NAXIS1']
    axissize2 = self.Head['NAXIS2']
    self.Head['NAXIS1']=axissize2
    self.Head['NAXIS2']=axissize1
    self.Head['history']="rotated 90 degrees from original for processing"
  def reflectAxis(self): #roll
    """Flip a frame about the vertical axis

    this allows the user to process all frames in a similar fasion by providing a way to orient all frames the same.
    """
    self.roll=0 if (self.roll) else 1
    print "reversing spatial direction"
    self.Data = np.fliplr(self.Data)
    self.Head['history']="spatial axis reversed for processing (assumes up down)"
  def switchDirection(self): #pitch
    """Flip a frame about the horizontal axis

    this allows the user to process all frames in a similar fasion by providing a way to orient all frames the same.
    """
    self.pitch=0 if (self.pitch) else 1
    print "reversing spectral direction"
    self.Data = np.flipud(self.Data)
    self.Head['history']="spectral axis reversed for processing (assumes up down)"


  def toFile(self,filename):
    """
    write this fits frame to a file
    """
    self.Head['history']="writing to file "+filename
    fitsdata=fits.PrimaryHDU(self.Data,header=self.Head)
    fitsdata.writeto(filename,clobber=True)

  def makeBiasFromOver(self,overscan,biasAxis,order=3): #todo: can use orientation and header to populate overscan and biasaxis, but not right now
    """
    generate a bias frame from the overscan region of this frame.

    user must supply overscan region boundaries(overscan) and position(biasAxis).  
    optionally the user can control the order of polynomial fit used in generating the bias frame
    """
    #use overscan to create a bias frame
    antibias = 0 if (biasAxis) else 1
    #print "DEBUG: making bias from overscan. overscan: ",overscan,"   biasAxis: ",biasAxis
    if (biasAxis==0):
      biasSection = np.array(self.Data[overscan[0]:overscan[1],:])
    elif (biasAxis==1): 
      biasSection = np.array(self.Data[:,overscan[0]:overscan[1]])
    else:
      raise Exception("illegal biasAxis specified in Frame.makeBiasFromOver.  Please specify 0 or 1")
    oneDBias=collapse2Dto1D(biasSection,biasAxis)
    fitParms = fit1D(oneDBias,order)
    recon1DBias = poly2oneD(fitParms,self.Data.shape[antibias])
    twoDBias = oneD2twoD(recon1DBias,self.Data.shape[biasAxis],biasAxis)
    #create a frame object and insert twoDBias as Data, set type as bias
    biasFrame = copy.deepcopy(self)
    biasFrame.type = 'bias'
    biasFrame.Data=twoDBias
    biasFrame.Head['history']="bias generated from overscan of "+str(self.originalFile)
    return biasFrame

    

#close class Frame


class interactiveLimitPick:
  """
  used for interactive spectral extraction.  Allows user to select points bounding intigration region and a number of sky regions.
  """
  def __init__(self,initialprofile,initialsmooth,spatialFitSpan):
    """
    user must provide profiles of the variation in the spectral direction, and the spatial direction span considered valid
    """
    self.initP = initialprofile
    self.initS = initialsmooth
    self.fitSpan = spatialFitSpan
    self.nextClick=0
    self.spectrumBounds = []
    self.skyBounds = []
    self.nextsky=[]
    self.lines={}
    plt.clf()
    self.fig=plt.figure()
    self.ax = self.fig.add_subplot(111)
    self.ax.set_title("Select integration window and sky windows")
    self.ax.plot(initialprofile,'c-')
    self.ax.axis((0,len(initialprofile),0,max(initialprofile[spatialFitSpan[0]:spatialFitSpan[1]])))
    #self.ax.plot(initialsmooth,'m-')
    self.ax.set_xlabel("spatial coordinate")
    self.ax.set_ylabel("intensity")
    self.ClickCID=self.fig.canvas.mpl_connect('button_press_event',self.onClicked)
    plt.show()
    print "please select integration limits for peak spectrum extraction"

  def pointInRegion(self,point):
    """
    is this point in a region we've already picked?

    return true if the point selected lies between two points that already define a region (integration or sky)
    """
    if (len(self.spectrumBounds) == 2):
      if (point > self.spectrumBounds[0] and point < self.spectrumBounds[1]):
        print "point lies in the spectral region"
        return True
    if (len(self.skyBounds)):
      print "checking skyBounds: ",self.skyBounds
      for p in self.skyBounds:
        print "this window: ",p
        if (point > p[0] and point < p[1]):
          print "point lies in a sky region"
          return True
    return False

  def onClicked(self,event):
    """
    handle click events for this class

    modifies state of internal variables to track whether a region is being opened or closed
    """
    #handle click events for region definition
    #ignore any clicks made while the window is in zoom mode
    if plt.get_current_fig_manager().toolbar.mode == 'zoom rect': return
    #ignore clicks outside the axes
    if event.inaxes!=self.ax.axes: return

    clickPix=int(event.xdata)
    # state 0 -> 1st integration limit
    # state 1 -> closing integration limit
    # state 2 -> 1st sky boundary
    # state 3 -> close sky boundary  (back to 2 when done, print close window reminder)

    #if user picks a line that is being used, have them try again.
    if (clickPix in self.lines.keys()):
      print "warning:  line already selected"
      return
    if (self.nextClick == 0):
      self.lines[clickPix] = self.ax.axvline(clickPix,color='r')
      self.spectrumBounds.append(clickPix)
      self.nextClick=1
      print "select other limit"
    elif (self.nextClick == 1):
      self.lines[clickPix] = self.ax.axvline(clickPix,color='r')
      self.spectrumBounds.append(clickPix)
      self.nextClick=2
      self.spectrumBounds.sort()
      print "select one or more sky regions"
    elif (self.nextClick == 2):
      if (self.pointInRegion(clickPix)):
        print "please select another point that doesn't lie in an existing selection"
        return
      self.lines[clickPix] = self.ax.axvline(clickPix,color='b')
      self.nextsky.append(clickPix)
      self.nextClick=3
      print "select other limit"
    elif (self.nextClick == 3):
      if (self.pointInRegion(clickPix)):
        print "please select another point that doesn't lie in an existing selection"
        return
      self.lines[clickPix] = self.ax.axvline(clickPix,color='b')
      self.nextsky.append(clickPix)
      self.nextClick=2
      self.skyBounds.append(sorted(self.nextsky))
      self.nextsky=[]
      print "please close window when done, or select an aditional sky region"
    else:
      die("something has gone terribly wrong in interactiveWindowPick")
    self.fig.canvas.draw()


  def getSpectralRegion(self):
    """
    return the bounds of the user selected spectral region
    """
    return self.spectrumBounds
  def getIgnoredRegions(self):
    """
    return a list of boundaries that define regions that should not be used for constructing a sky image
    """
    print "DEBUG ignoredRegions:  skybounds as built:",self.skyBounds
    print "DEBUG ignoredRegions:  spectrumbounds    :",self.spectrumBounds
    sb = self.skyBounds
    #sb.append(self.spectrumBounds)
    bounds = sorted(sb, key=lambda first: first[0])
    ignores = [[0,max(self.fitSpan[0],bounds[0][0])]]
    #loop appending
    for i in range(1,len(bounds)):
      ignores.append([bounds[i-1][1],bounds[i][0]])
    ignores.append([min(self.fitSpan[1],bounds[-1][1]),len(self.initP)])    # need fitSpan and shape in pick
    return ignores


#close class interactiveLimitPick

def simpleExtract(body):
  """
  naive extraction, just sum all columns.

  do any trimming before calling simpleExtract
  """
  spectrum=np.sum(body,1)
  if (testing_display and False):     #this block is for very simple debugging
    plt.title("Initial Standard Spectrum")
    plt.ylabel("intensity?")
    plt.xlabel("spectral pixel coordinate")
    plt.plot(spectrum,'b-')
    plt.ylim([0,max(spectrum)])
    plt.savefig("debug/standardSpectrum.png")
    plt.show()
    plt.clf()
  return spectrum

def radiansFromString(rawAngle):
  """
  return radians from a string representing the angle

  this is mainly used for dealing with angles stored in fits headers.  As there is not a real standard for this,
  this is a place where future adaptation could be required (hour angles, varied formats, etc...)
  """
  #degIndex = rawAngle.find('deg')
  thetadeg=0
  if ('deg' in rawAngle):
    parts=rawAngle.split('deg')
    thetadeg=thetadeg+float(parts[0])
    if ('min' in parts[1]):
      minparts=parts[1].split('min')
      minutes = minparts[0]
      thetadeg=thetadeg+float(minutes)/60
  rads=thetadeg*math.pi/180
  return rads

def gratingFromString(rawGrate):
  """
  get the grating from a string

  this is mainly used for dealing with information from fits headers.  As there is not a real standard for this,
  this is a place where future adaptation could be required
  """
  if ('/' in rawGrate):
    parts = rawGrate.split('/')
    return float(parts[0])
  return float(rawGrate)

def computeDispersion(fixedAngle,theta,pixelSize,focalLength,grating):
  """
  compute the dispersion from the instrument parameters
  """
  return 10**4*math.cos(fixedAngle-theta)*pixelSize/(focalLength*grating)
def computeCentralWL(theta,grate,fixedAngle):
  """
  compute the central wavelength of the observation
  """
  d=grate/10**7
  k1=1+math.cos(fixedAngle)
  k2=math.sin(fixedAngle)
  b=-2.*k1*math.sin(theta)/k2
  c = (1 + (k1/k2)**2)*(math.sin(theta))**2 - 1.0
  k3 = 0.5*(-1.0*b - math.sqrt(b**2 - 4*c))
  lam = k3*k2/d
  return lam

#def computeCentralWL(theta,grate):
#  """
#  given a central wavelength and a grating, compute the central wavelength
#  """
#  return 10**7*math.sin(theta)/grate
def initialCalibration(dispersion,centralLambda,size):
  """
  give a preliminary calibration

  use the dispersion and the central wavlength to fill an array of the size of our spectrum with a computed approximate wavelength
  """
  #print "calibration input vars:  center:",centralLambda,"   dispersion:",dispersion,"    size",size
  #print " offset:",dispersion*size/2
  return np.linspace(centralLambda-dispersion*size/2,centralLambda+dispersion*size/2,size)

def seekMin(func,startpt,step=1):
  """
  given a vector, and a point in the vector, march in a direction until a local minimum is found
  """
  pointer=startpt
  while(1<pointer<len(func)-1):
    prevpt = pointer
    pointer+=step
    if (func[pointer]>func[prevpt]):
      return prevpt
  if (step>0):
    return len(func)-1
  else:
    return 0

def seekMax(func,startpt,step=1):
  """
  given a vector, and a point in the vector, march in a direction until a local maximum is found
  """
  pointer=startpt
  while(1<pointer<len(func)-1):
    prevpt = pointer
    pointer+=step
    if (func[pointer]<func[prevpt]):
      return prevpt
  if (step>0):
    return len(func)-1
  else:
    return 0

def gaus(x,a,x0,sigma,c):
    return a*np.exp(-(x-x0)**2/(2*sigma**2))+c

def gaussPeak(func,peak):
  """
  given a vector and an approximate center of a peak, find the true center of that peak assuming a gaussian
  """
  #find bounding local minimums to use for initial sigma extimation
  Lmin=seekMin(func,peak,-1)
  Rmin=seekMin(func,peak,1)
  #if a peak is at a minimum (makes no sense) offset bound by one.
  if (peak == Lmin):
    Lmin-=1
  if (peak == Rmin):
    Rmin+=1
  if (USERDefs.debug):
    print "DEBUG: checking sanity global, Lmin:",Lmin,"  Rmin:",Rmin,"   peak:",peak
  offset=Lmin
  localFunc = func[Lmin:Rmin]
  peak=peak-Lmin
  Lmin=0
  Rmin=len(localFunc)-1
  if (USERDefs.debug):
    print "DEBUG: checking sanity local, Lmin:",Lmin,"  Rmin:",Rmin,"   peak:",peak,"  len:",len(localFunc)
  Lhalf=(localFunc[peak]-localFunc[Lmin])/2
  Lfunc=[x-Lhalf for x in localFunc]
  Lfwhm=seekZero(Lfunc,peak,-1)
  Rhalf=(localFunc[peak]-localFunc[Rmin])/2
  Rfunc=[x-Rhalf for x in localFunc]
  Rfwhm=seekZero(Rfunc,peak,1)
  FWHM = Rfwhm-Lfwhm
  sigFWHM=FWHM/2.35
  #print "DEBUG sigGuess: ",sigGuess,"  sigFWHM: ",sigFWHM
  sigGuess=sigFWHM
  constGuess=(localFunc[Rmin]+localFunc[Lmin])/2
  x=range(len(localFunc))
  #print "DEBUG: initial parameters for curve fit: ",func[peak]," ",peak," ",sigGuess," ",constGuess
  try:
    popt, pcov=scipy.optimize.curve_fit(gaus,x,localFunc,p0=[localFunc[peak],peak,sigGuess,constGuess])
  except:
    warnings.warn("WARNING subpixel peak refinement failed at pixel "+str(peak+offset)+" using unrefined position",UserWarning)
    return peak+offset
  #print "DEBUG:  optimum gauss params: ",popt
  #print "DEBUG:  evaluation of initial gauss at 5 points: Lmin:",gaus(Lmin,func[peak],peak,sigGuess,constGuess)," Lhalf:",gaus((Lmin+peak)/2,func[peak],peak,sigGuess,constGuess)," peak:",gaus(peak,func[peak],peak,sigGuess,constGue          ss)," Rhalf:",gaus((Rmin+peak)/2,func[peak],peak,sigGuess,constGuess)," Rmin:",gaus(Rmin,func[peak],peak,sigGuess,constGuess)
  g_peak=popt[1]
  if (abs(popt[1]-peak) > sigGuess):
    warnings.warn("WARNING gross peak fitting (pixel) and refined (gauss) have a substantial difference.  gauss="+str(g_peak+offset)+"  gross="+str(peak+offset)+".",UserWarning)
    #if they varry wildly, it's probably the gauss that is further off, because it could have traveled in ways the normal peak cannot.
    return peak+offset
  return g_peak+offset

def gaussPeak2(func,peak):
  """
  given a vector and an approximate center of a peak, find the true center of that peak assuming a gaussian
  """
  #find bounding local minimums to use for initial sigma extimation
  Lmin=seekMin(func,peak,-1)
  Rmin=seekMin(func,peak,1)
  localFunc = copy.deepcopy(func)
  for x in range(0,Lmin):
    localFunc[x]=Lmin
  for x in range(Rmin,len(func)):
    localFunc[x]=Rmin
  #localFunc[0:Lmin] = func[Lmin]
  #localFunc[Rmin:-1] = func[Rmin]
  Lhalf=(func[peak]-func[Lmin])/2
  Lfunc=[x-Lhalf for x in func]
  Lfwhm=seekZero(Lfunc,peak,-1)
  Rhalf=(func[peak]-func[Rmin])/2
  Rfunc=[x-Rhalf for x in func]
  Rfwhm=seekZero(Rfunc,peak,1)
  FWHM = Rfwhm-Lfwhm
  sigFWHM=FWHM/2.35
  #print "DEBUG sigGuess: ",sigGuess,"  sigFWHM: ",sigFWHM
  sigGuess=sigFWHM
  constGuess=(func[Rmin]+func[Lmin])/2
  x=range(len(func))
  #print "DEBUG: initial parameters for curve fit: ",func[peak]," ",peak," ",sigGuess," ",constGuess
  popt, pcov=scipy.optimize.curve_fit(gaus,x,localFunc,p0=[localFunc[peak],peak,sigGuess,constGuess])
  #print "DEBUG:  optimum gauss params: ",popt
  #print "DEBUG:  evaluation of initial gauss at 5 points: Lmin:",gaus(Lmin,func[peak],peak,sigGuess,constGuess)," Lhalf:",gaus((Lmin+peak)/2,func[peak],peak,sigGuess,constGuess)," peak:",gaus(peak,func[peak],peak,sigGuess,constGue     ss)," Rhalf:",gaus((Rmin+peak)/2,func[peak],peak,sigGuess,constGuess)," Rmin:",gaus(Rmin,func[peak],peak,sigGuess,constGuess)
  g_peak=popt[1]
  if (abs(popt[1]-peak) > 1):
    warnings.warn("WARNING gross peak fitting (pixel) and refined (gauss) have a substantial difference.  gauss="+str(g_peak)+"  gross="+str(peak)+".",UserWarning)
    #if they varry wildly, it's probably the gauss that is further off, because it could have traveled in ways the normal peak cannot.
    return peak
  return g_peak

def gaussPeakOrig(func,peak):
  """
  given a vector and an approximate center of a peak, find the true center of that peak assuming a gaussian
  """
  #find bounding local minimums to use for initial sigma extimation
  Lmin=seekMin(func,peak,-1)
  Rmin=seekMin(func,peak,1)
  #TODO try replacing all values below Lmin with value at Lmin, and above Rmin...
  #print "DEBUG: bounds: ",Lmin,",",Rmin,"  peak:",peak
  #sigtot=0
  #for i in range(Lmin,Rmin):
  #  sigtot+=((i-peak)*(func[i]-func[peak])**2)
  #sigGuess=math.sqrt(sigtot/(Rmin-Lmin))
  #none  of the stuff I have tried for computing a stddev from the function is working.  Try FWHM then back out sigma...
  Lhalf=(func[peak]-func[Lmin])/2
  Lfunc=[x-Lhalf for x in func]
  Lfwhm=seekZero(Lfunc,peak,-1)
  Rhalf=(func[peak]-func[Rmin])/2
  Rfunc=[x-Rhalf for x in func]
  Rfwhm=seekZero(Rfunc,peak,1)
  FWHM = Rfwhm-Lfwhm
  sigFWHM=FWHM/2.35
  #print "DEBUG sigGuess: ",sigGuess,"  sigFWHM: ",sigFWHM
  sigGuess=sigFWHM
  constGuess=(func[Rmin]+func[Lmin])/2
  x=range(len(func))
  #print "DEBUG: initial parameters for curve fit: ",func[peak]," ",peak," ",sigGuess," ",constGuess
  popt, pcov=scipy.optimize.curve_fit(gaus,x,func,p0=[func[peak],peak,sigGuess,constGuess])
  #print "DEBUG:  optimum gauss params: ",popt
  #print "DEBUG:  evaluation of initial gauss at 5 points: Lmin:",gaus(Lmin,func[peak],peak,sigGuess,constGuess)," Lhalf:",gaus((Lmin+peak)/2,func[peak],peak,sigGuess,constGuess)," peak:",gaus(peak,func[peak],peak,sigGuess,constGuess)," Rhalf:",gaus((Rmin+peak)/2,func[peak],peak,sigGuess,constGuess)," Rmin:",gaus(Rmin,func[peak],peak,sigGuess,constGuess)
  g_peak=popt[1]
  if (abs(popt[1]-peak) > 1):
    warnings.warn("WARNING gross peak fitting (pixel) and refined (gauss) have a substantial difference.  gauss="+str(g_peak)+"  gross="+str(peak)+".",UserWarning)
    #if they varry wildly, it's probably the gauss that is further off, because it could have traveled in ways the normal peak cannot.
    return peak
  return g_peak

def seekZero(func,startpt,step=1):
  """
  given a vector, and a point in the vector, march in a direction until a zero crossing is found
  """
  signs = np.sign(func)
  pointer=startpt
  while (1<pointer<len(func)-1):
    pointer+=step
    if (signs[pointer]*signs[startpt] == -1):
      return pointer
  #return -1
  if (step>0):
    return len(func)-1
  else:
    return 0


def loadLamps(inlist):   
  """
  load reference lamp files
  """
  #todo: should make it handle single file input as well
  peaks=[]   
  for f in inlist:
    file=open(f,'r')
    for l in file:
      if (l[0].isdigit()):
        peaks.append(float(l.split()[0]))
  peaks.sort()
  return peaks


def calibrateWavelength(lampPeaks,refPeaks,initialGuessCalibration,correlation_cluster_size=5,matching_threshold=5):
  """
  find the offset between our reference peaks and our extracted lamp image peaks

  this is done by finding the maximum cross correlation between synthetic spectra made of subsets of the lamp
  emission lines and the reference emission lines.

  return value is a list of pairs of lamp lines (in pixel coordinates) and the reference line believed to match

  lampPeaks - list of peaks extracted from lamp images inpixel coordinates
  refPeaks  - list of referance peaks for the lamp in angstroms
  initialGuessCalibration - initial guess at mapping from pixel space to angstroms.
  correlation_cluster_size - number of tightly packed peaks to use in the calibration.
  matching_threshold - maximum allowed difference (in angstroms) between a lamp peak 
    calibration attempt and the nearest ref peak for accepting them as being the same peak

  using initialGuessCalibration, find the offset between a the calibrated lampPeaks, and the refPeaks
    return dict of lamp peaks in pixel space:wl from reference
  """
  #as input the lampPeaks are in order of prominance.  sort by WL
  S_peaks = sorted(lampPeaks)
  #use a tight cluster of a small number of the peaks to correlate for initial shift
  #distance to the corelation_cluster_sizeth higher peak for first n-correlation_cluster_size peaks
  if (correlation_cluster_size >= len(S_peaks)-1):
    print "using all peaks"
    Corr_peaks = S_peaks
  else:
    print "using a cluster of ",correlation_cluster_size," peaks"
    dist_n=[]
    for i in range(len(S_peaks)-correlation_cluster_size):
      dist_n.append(S_peaks[i+correlation_cluster_size] - S_peaks[i])
    #idx, val = min(enumerate(dist_n), key=operator.itemgetter(1))
    val, idx = min((val, idx) for (idx, val) in enumerate(dist_n))
    print "the pixel span for the minimum ",correlation_cluster_size," peaks is ",val," which starts from the ",idx," peak of ",len(S_peaks)," calibration peaks"
    Corr_peaks = S_peaks[idx:idx+correlation_cluster_size]
  #debug info
  print "pixel space listing of correlation peaks: ",Corr_peaks
  print "calibration guess for initial correlation peaks\n",initialGuessCalibration[Corr_peaks]

  #create snythtic spectra for the lamp peaks and for the ref peaks
  # the synthetic spectra will have the same size, and will go from 0 to the max 
  #    value present in either the preliminary calibration or the reference (plus a fudge)
  corrSpecSize = int(max([max(refPeaks),initialGuessCalibration[Corr_peaks[-1]]]))+5  
  refSpec = np.zeros(corrSpecSize)
  lampSpec = np.zeros(corrSpecSize)
  for p in refPeaks:
    i_part = int(p)
    f_part = p-i_part
    refSpec[i_part]+=1-f_part
    refSpec[i_part+1]+=f_part
  for p in initialGuessCalibration[Corr_peaks]:
    i_part = int(p)
    f_part = p-i_part
    lampSpec[i_part]+=1-f_part
    lampSpec[i_part+1]+=f_part

  #we should have synthetic spectra for referance and lamp peaks at this point.
  #  things to try:  correlate.   Smooth then correlate, so peaks are less discrete and initial dispersion error is forgiven
  corr = np.correlate(refSpec,lampSpec,'full')
  maxCorr= np.argmax(corr)
  offset = corrSpecSize-maxCorr
  #offset = maxCorr-corrSpecSize
  print "maximum correlation occurs at a shift of ",offset
  print "initial calibration with shift applied gives cluster:",initialGuessCalibration[Corr_peaks]+offset

  #Try a different method for getting the Offset
  #offset = findCalibrationOffset(initialGuessCalibration[Corr_peaks],refPeaks)
  #print "offset computed by alternate method: ",offset
  #print "initial calibration with alternate shift applied gives cluster:",initialGuessCalibration[Corr_peaks]+offset
  peakPairs = {}
  for lampPeak in Corr_peaks:
    calibrated = initialGuessCalibration[lampPeak]+offset
    #find closest reference peak
    line=refPeaks[min(range(len(refPeaks)), key=lambda i: abs(refPeaks[i]-calibrated))]
    #print "DEBUGGING: checking proximity of calibrated lamp peak, ",calibrated," and reference peak, ",line
    if (abs(calibrated-line) <= matching_threshold):
      peakPairs[lampPeak] = line
  return peakPairs

def findSkyCalOffset(initCal,skyspec,reflines,span,maxinitialdelta=3):
  skypeaks=findPeaks(skyspec,spectralFitSpan=span)
  #get closest skypeak to each refline, disregard if too far away
  peakpairs={}
  for r in reflines:
    peak = skypeaks[min(range(len(skypeaks)), key=lambda i: abs(skypeaks[i]-r))]
    delta=r-peak
    if (abs(delta)<maxinitialdelta):
      if (peak not in peakpairs):
        peakpairs[peak]=[]
      peakpairs[peak].append(delta)
  mindeltas=[]
  for peak in peakpairs:
    mindeltas.append(min(peakpairs[peak]))
  offset = sum(mindeltas)/float(len(mindeltas))
  return offset
    
    

def findCalibrationOffset(calSet,refSet,maxOffset=0,stepSize=0):
  """
  find the offset value that minimizes the difference between calSet and refSet
  """
  calSet.sort()
  refSet.sort()
  if stepSize is 0:   #need to compute a reasonable stepSize from the refSet.  Hueristic: 1/4 of minimum dif between reference peaks, not to leave the range of .1 to 1 
    deltas = [abs(refSet[i]-refSet[i+1]) for i in range(len(refSet)-1)]
    stepSize = max(.1,min(min(deltas)/4,1))
  if maxOffset is 0:  # use the entire span
    lowEnd = abs(refSet[0]-calSet[0])
    highEnd= abs(refSet[-1]-calSet[-1])
    maxOffset = max(lowEnd,highEnd)
  print "DEBUG: stepSize: ",stepSize," and maxoffSet:" ,maxOffset,"   Total steps: ",2*maxOffset/stepSize
  #fill an array with the RMS difference between the offset calibration set and the closest reference peaks
  corrValues=[]
  for offset in np.arange(-maxOffset,maxOffset,stepSize):
    offsetSet = [c+offset for c in calSet]
    val = closestPeakDiff(offsetSet,refSet)
    corrValues.append(val)
    print "DEBUGGING: offset:",offset,"    SSvalue:",val
  #offsetRet = corrValues.index(min(corrValues))*stepSize
  #offsetRet = min(xrange(len(corrValues)),key=corrValues.__getitem__)*stepSize
  minVal = min(corrValues)
  offsetRet = -maxOffset+np.argmin(corrValues)*stepSize
  print "DEBUG: the min value is ",minVal
  print "DEBUG: returning:",offsetRet
  return offsetRet


def closestPeakDiff(calSet,refSet):
  """
  for each item in calSet, find the closest item in refSet.  return the squared sum of the delta between all of the calSet members and their
  refSet partners
  """
  #because John can't figure out how to use the crosscorrelation he will try minimizing this for all shifts to get automatic calibration working
  sumSquare = 0
  for calibrated in calSet:
    refPartner = refSet[min(range(len(refSet)), key=lambda i: abs(refSet[i]-calibrated))]
    delta = calibrated-refPartner
    sumSquare+=delta**2
    #print "DEBUGGING: calibrated:",calibrated," refPartner:",refPartner,"  delta:",delta,"       sumsSquare:",sumSquare
  return sumSquare
  

def is_number(s):
  try:
    float(s)
    return True
  except ValueError:
    return False

def interactiveCalibrationCollector(lampSpec,framepitch=0,refPeaks=None):
  """
  via user interaction, collect pairs of pixel space peak -> wavelength space peak
  """
  #collect a series of pixel coordinate->wavelength pairs from user input.  
  #  User clicks near a peak, the actuall max location is found, and user inputs a wavelength for that peak
  class WLpeakPicker:
    def __init__(self,lampSpec,framepitch=0,refPeaks=None):
      self.LampSpec=lampSpec
      #if (framepitch==1):
      #  self.LampSpec=self.LampSpec[::-1]
      self.peaks={}
      self.lines={}
      self.refPeaks=refPeaks
      plt.clf()
    def show(self):
      self.fig=plt.figure()
      self.ax = self.fig.add_subplot(111)
      self.ax.set_title("click to select calibration peaks")
      self.ax.plot(lampSpec)
      self.ax.set_xlabel("pixel coordinate")
      self.ax.set_ylabel("intensity")
      maxv=max(self.LampSpec)
      #get xlims from left and right by looking for threshold value
      x0=0
      x1=len(self.LampSpec)-1
      while(self.LampSpec[x0]<maxv*.01):
        x0+=1
      while(self.LampSpec[x1]<maxv*.01):
        x1+=-1
      self.ax.set_ylim(0,maxv)
      self.ax.set_xlim(x0,x1)
      self.ClickCID=self.fig.canvas.mpl_connect('button_press_event',self.onClicked)
      #self.fig.canvas.draw()
      plt.show()
    def onClicked(self,event):
      if plt.get_current_fig_manager().toolbar.mode == 'zoom rect':
        print "We are in zoom with Rectangle - mode"
        return
      if event.inaxes!=self.ax.axes: return
      #print "xpixelcoord is",event.xdata
      clickPix=int(event.xdata)
      if (self.LampSpec[clickPix-1]>self.LampSpec[clickPix] and self.LampSpec[clickPix+1]<self.LampSpec[clickPix]):
        #go left
        peak=seekMax(self.LampSpec,clickPix,-1)
      elif (self.LampSpec[clickPix-1]<self.LampSpec[clickPix] and self.LampSpec[clickPix+1]>self.LampSpec[clickPix]):
        #go right
        peak=seekMax(self.LampSpec,clickPix,1)
      elif (self.LampSpec[clickPix-1]<self.LampSpec[clickPix] and self.LampSpec[clickPix+1]<self.LampSpec[clickPix]):
        #max clicked
        peak=clickPix
      else:
        #min clicked or problem
        left=seekMax(self.LampSpec,clickPix,-1)
        right=seekMax(self.LampSpec,clickPix,1)
        if (self.LampSpec[left] > self.LampSpec[right]):
          peak = left
        else:
          peak=right
      peak=gaussPeak(self.LampSpec,peak)
      print "peak refined coordinate=",peak
      if (peak not in self.lines.keys()):
        self.lines[peak] = self.ax.axvline(peak,color='r')
      #plt.show()
      self.fig.canvas.draw()
      print "peak pixel coordinate=",peak
      #insert subpixel peak finder here
      textWL= raw_input("enter wavelength in angstroms (or cancel, or delete):")
      print "getting a WL from ",textWL
      if (not is_number(textWL)):
        l = self.lines[peak];
        l.remove()
        del l
        del self.lines[peak]
        self.fig.canvas.draw()
        if (textWL.lower() == 'delete' or textWL.lower() == 'remove' or textWL.lower() == 'del' or textWL.lower() == 'd'):
          print "deleting at peak ",peak
          if (peak in self.peaks.keys()):
            del self.peaks[peak]
          else:
            print "no peak found at coordinate ",peak,", nothing deleted"
        return
      WL=float(textWL)
      if (self.refPeaks):
        if (WL not in self.refPeaks):
          #entered WL is not in the reference set.  Find the closest one and ask the user what to do.
          #refindex = min(range(len(localRef)), key=lambda i: abs(localRef[i]-calibration[p]))
          closestRef = self.refPeaks[min(range(len(self.refPeaks)), key=lambda i: abs(self.refPeaks[i]-WL))]
          action= raw_input("ALERT: User wavelength: "+str(WL)+" is not in the reference set.  Closest reference peak is "+str(closestRef)+
            "(delta="+str(abs(closestRef-WL))+"). Which would you like to use? (enter 'user', 'ref', or 'cancel'):")
          if (action.lower() == 'user' or action.lower() == 'u'):
            print "using ",WL
            self.peaks[peak] = WL
          if (action.lower() == 'reference' or action.lower() == 'ref'  or action.lower() == 'r'):
            print "using ",closestRef
            self.peaks[peak] = closestRef
          return
      else:
        print "DEBUG: I don't have a reference peak set, why?"
      self.peaks[peak]=WL
    def getPeaks(self):
      print "returning calibration peaks ",self.peaks
      ret = self.peaks
      #ret=[]
      #for k in sorted(self.peaks.keys()):
      #  ret.append([k,self.peaks[k]])
      ##print "returning these peaks ",ret
      return ret
  #close class
  calibrator = WLpeakPicker(lampSpec,framepitch=0,refPeaks=refPeaks)
  calibrator.show()
  peakPairs=calibrator.getPeaks()
  return peakPairs


def pairs2Calibration(pairDict,size,maxorder=3):
  """
  fit a polynomial through the peaks in the calibration set, give back an array of the resulting calibration
  """
  #print "DEBUG pair2cal incoming state: ",pairDict
  fitx=[pixel for pixel in pairDict]
  fitx.sort(key=float)
  fity=[float(pairDict[pix]) for pix in fitx]
  fitx=[float(x) for x in fitx]
  #print "DEBUG pair2cal prefit state: "
  #print "      X:",fitx
  #print "      Y:",fity
  #print "  order:",min(len(pairDict)-2,maxorder)
  calfit = fit1DNonuniform(fity,fitx,order=min(len(pairDict)-2,maxorder))
  #calfit = fit1DNonuniform(fity,fitx,order=1)
  #print "DEBUG:  fit as: ",calfit
  poly = poly2oneD(calfit,size)
  #print "DEBUG LOOP: sanity check pairs->calibration"
  #for p in fitx:
  #  print "DEBUG: pixel: ",p," ref: ",pairDict[p]," calibrated: ",poly[p]
  return poly

def expandCalibration(initialPairs,foundPeaks,referencePeaks,order=3,matching_threshold=1):
  """
  incorporate more lamplines and referencepeaks into the calibration set

  take a set of peaks in pixel space and their mapping to wavelength space, 
    and incorporate additional pixel->wavelength pairs using polynomial 
    fitting  andproximity based matching
  initialPairs - dictionary of pixel->WL
  foundPeaks - peaks in pixel space
  referencePeaks - possible peaks in WL space
  order - max order of polynomial used for fitting peaks
  maxDiff - maximum allowed difference in WL space between a peak's predicted location from the 
    current calibration and a referance peak for including them as a pair in the expanded calibration
  """

  #create the pair set we will build
  expandedPairs = initialPairs
  #create local versions of the peak lists that will be emptied as the expansion continues
  localRef = sorted(referencePeaks)
  remainingPeaks = sorted(foundPeaks)
  #remove what's already been matched
  for pixel in initialPairs:
    if (initialPairs[pixel] in localRef):
      localRef.remove(initialPairs[pixel])
    else:
      #TODO add to this with something that checks for stuff obscenely close to the reference peak
      warnings.warn("Wavelength ("+str(initialPairs[pixel])+") in initial calibration set is absent from reference set",UserWarning)
    if (pixel in remainingPeaks):
      remainingPeaks.remove(pixel)
    else:
      warnings.warn("Peak in initial calibration set is absent from calibration set",UserWarning)

  #incorporate additional pairs as possible
  spanmult=1 #see heuristic explanation below
  direction=0 #try to alternate search directions when growing the calibration region.
              #  This should prevent one side from pulling the other off the correct solution.
  while (remainingPeaks):
    #print "DEBUG: current calibration:",expandedPairs
    #fit a polynomial through the peaks currently in the calibration
    #fitx=[pixel for pixel in expandedPairs]
    #fity=[expandedPairs[pixel] for pixel in expandedPairs]
    #fitx.sort()
    #fity.sort()
    #calfit = fit1DNonuniform(fity,fitx,order=max(len(expandedPairs)-2,3))
    #calibration = poly2oneD(calfit,max(remainingPeaks)+1)
    calibration = pairs2Calibration(expandedPairs,int(max(foundPeaks)+1))

 
    #Favor peaks close to the peaks already in the calibration so that extrapolation error is minimized.
    #  heuristic: look first for peaks within a 'calibration span' of the existing calibration set and use the one closest to a referance peak
    #  grow region if no suitible peaks are found.  Return when no suitible peaks are found and search region encompasses all peaks
    pix = [pixel for pixel in expandedPairs]
    minPix = min(pix)
    maxPix = max(pix)
    span = maxPix-minPix
    searchMin = minPix
    searchMax = maxPix
    inc = spanmult*span
    if (direction <=0):
      searchMin = searchMin-inc
    if (direction >=0):
      searchMax = searchMax+inc
    remainingNP = np.array(remainingPeaks)
    candidates = remainingNP[(remainingNP>=searchMin) & (remainingNP<=searchMax)]
    if (len(candidates)==0):
      spanmult+=.5
      continue 
    #print "DEBUG: searching for additional peaks.  Have ",len(candidates)," candidates in the range of ",searchMin," to ",searchMax,"."
    #build list of calibrated pixel->nearest reference deltas
    deltas=[]
    #for p in remainingPeaks:
    for p in candidates:
      refindex = min(range(len(localRef)), key=lambda i: abs(localRef[i]-calibration[p]))
      refpeak = localRef[refindex]
      delta = abs(refpeak-calibration[p])
      #print "DEBUG: peak ",p," calibrates to ",calibration[p],".  The closest reference peak is ",refpeak,".  delta:",delta,"."
      deltas.append(delta)
    bestPeakIndex = deltas.index(min(deltas))
    if (deltas[bestPeakIndex] <= matching_threshold):
      #incorporate this peak in calibration set and repeat
      peak = candidates[bestPeakIndex]
      refindex = min(range(len(localRef)), key=lambda i: abs(localRef[i]-calibration[peak]))
      refpeak = localRef[refindex]
      expandedPairs[peak] = refpeak
      remainingPeaks.pop(remainingPeaks.index(peak))
      localRef.pop(localRef.index(refpeak))
      spanmult=1
      print "DEBUG: added ",peak,":",refpeak," to the calibration set.  Error in preaddition calibration for this peak was ",deltas[bestPeakIndex],"."
      if ((peak < maxPix and peak > minPix) or searchMax > max(remainingPeaks) or searchMin < min(remainingPeaks) ):
        print "DEBUG: searching both ways"
        direction = 0
      elif (peak <= minPix):
        print "DEBUG:  searching up only"
        direction = 1
      elif (peak >= maxPix):
        print "DEBUG:  searching down only"
        direction = -1
      else:
        raise RuntimeError('a logical branch thought to be excluded was executed in SpecReduceUtils.expandCalibration')
    else:  #no suitible peaks found at this time
      if (len(candidates) == len(remainingPeaks)):
        #all peaks were considered, this is as good as we are gonna do
        print "DEBUG: ",len(remainingPeaks)," peaks were not used in calibration.  If this number is large, it could represent a problem"
        return expandedPairs
      else:  #there are peaks outside of the range we considered.  Take a risk and increase extrapolation allowance
        spanmult+=.5
        print "DEBUG: no matches found in extrapolation region, try a larger region.  spanmult:",spanmult,"  direction:",direction
        #if the search has grown beyond the span, it will only have been on one end, so reset spanmult and set direction to neutral
        if (searchMax > max(remainingPeaks) or searchMin < min(remainingPeaks)):
          if (direction != 0):
            spanmult=1
          direction=0
          print "DEBUG: search reached a limit, searching both directions.  spanmult:",spanmult,"  direction:",direction
  print "DEBUG: all peaks were used in the calibration"
  return expandedPairs



def findPeaks(input,spectralFitSpan=None,enoughpeaks=50):
  """
  find prominant peaks in an array
  """
  if (spectralFitSpan is None): spectralFitSpan=[300,1700]
  #debug
  #display=True
  fit = fit1D(input,3,spectralFitSpan)
  fitted = poly2oneD(fit,len(input))
  fittedc = [f if f>0 else -1 for f in fitted]
  fittedc = [f if f>10000 else 10000 for f in fittedc]
  subtracted=input-fittedc
  peaks = []
  while (len(peaks) < enoughpeaks):
    maxX=np.argmax(subtracted)
    if (subtracted[maxX]<=0):
      break
    if (maxX<=0 or maxX>=len(subtracted)-1):
      print "DEBUG findPeaks has found a peak at a margin.  What's it look like now?  ",subtracted
      #print "DEBUG  all elements >0"
      #for x in subtracted:
      #  if (x > 0):
      #    print x
      if maxX>=len(subtracted)-1:
        print "removing right bound maximum"
        subtracted[maxX]=-1
      if maxX<=0:
        print "removing left bound maximum"
        subtracted[0]=-1
      continue
    intensity=input[maxX]
    bounds = [seekZero(subtracted,maxX,-1),seekZero(subtracted,maxX,1)]
    #print "DEBUG: bounds of replacement: ",bounds, " for peak at ",maxX
    subtracted[bounds[0]:bounds[1]]=-1*np.ones(bounds[1]-bounds[0])
    refinedPeak=gaussPeak(input,maxX)
    peaks.append(refinedPeak)
    #print "DEBUG: pixelpeak: ",maxX," fitted:",refinedPeak
    #print "we currently have "+str(len(peaks))+" peaks."
  print str(len(peaks))+" peaks found"
  print peaks
  #debug
  if (False):
    plt.title("debugging specReduceUtils.findPeaks plot of spectrum with found peak locations")
    plt.plot(input)
    for p in peaks:
      plt.axvline(p,color='r')
    plt.show()
    plt.clf()
  return peaks

def frameInCommon(frmlist):
  """
  create a new frame that has properties shared by all frames in frmlist (useful for operations that combine frames)
  """
  newFrame = Frame()
  f1 = frmlist.pop(0)
  rollsame=True
  pitchsame=True
  yawsame=True
  typesame=True
  orientsame=True
  #cards1 = copy.deepcopy(f1.Head.cards)
  newFrame.Head = copy.deepcopy(f1.Head)
  cards1 = newFrame.Head.cards
  for f in frmlist:
    rollsame = rollsame and f1.roll == f.roll
    pitchsame = pitchsame and f1.pitch == f.pitch
    yawsame = yawsame and f1.yaw == f.yaw
    typesame = typesame and f1.type == f.type
    orientsame=orientsame and f1.frameSetup == f.frameSetup
    for c in cards1:
      keyword = c[0]
      #print "working on keyword: ",keyword
      try:
        f.Head[keyword]
      except KeyError:
        del newFrame.Head[keyword]
        #print "deleting keyword that is not universally present:",keyword
      else:
        del newFrame.Head[keyword]
        #print "deleting keyword that doesn't match:",keyword
  if (rollsame):
    newFrame.roll=f1.roll
  if (pitchsame):
    newFrame.pitch=f1.pitch
  if (yawsame):
    newFrame.yaw=f1.yaw
  if (typesame):
    newFrame.type=f1.type
  if (orientsame):
    newFrame.frameSetup=f1.frameSetup
  return newFrame


def frameMedian(frmlist):
  """
  create a new frame that has Data = median of Data for all input frames.
  """
  newFrame = frameInCommon(frmlist) 
  datas=[]
  for f in frmlist:
    datas.append(f.Data)
  newFrame.Data=np.median(datas,0)
  newFrame.Head['history'] = str(newFrame.type)+" created as median of multiple fits"
  return newFrame

def frameDivision(num,denom):
  """
  create new frame that has Data = num.Data/denom.Data
  """
  newFrame=frameInCommon([num,denom])
  newFrame.Data = num.Data/denom.Data
  newFrame.Head['history'] = str(newFrame.type)+" created from division of multiple fits"
  return newFrame

def frameSubtract(first,last):
  """
  create a new frame that has Data= first.Data-second.Data
  """
  newFrame=frameInCommon([first,last])
  newFrame.Data=first.Data-last.Data
  newFrame.Head['history'] = str(newFrame.type)+" created from difference of 2 fits"
  return newFrame

def removeCosmics(inframe):
  """
  remove cosmic rays from the frame
  """
  cosmicFrm = copy.deepcopy(inframe)
  cosmicMsk = copy.deepcopy(inframe)
  cosImage = cosmics.cosmicsimage(inframe.Data)
  cosImage.run()
  numMask = np.cast["uint8"](cosImage.mask)
  cosmicMsk.Data = numMask
  cosmicFrm.Head['history']= "decosmiced with LAcosmic"
  cosmicMsk.Head['history']= "cosmic ray mask from LAcosmic"
  cosmicFrm.Data = cosImage.cleanarray
  return cosmicFrm,cosmicMsk

def optimalExtractWrapper(inframe,frameSetup,interactive=False,name="frameNameNotSpecified",calibration=None,telluric=None,fluxCal=None,tracefile=None,aboveSky=.10,savePath=USERDefs.savepath):
  """
  perform Horne's extraction on all significant, or user selected, spectra on the frame

  inframe is the frame we are working with.  frameSetup is a collection of parameters pertaining to that frame, interactive: should the user pick regions
    calibration is the wavelength calibration, tracefile lets you read a trace from a file rather than computing one, and aboveSky is a 
    sensitivity parameter for non-interactive peak finding (higher is less sensitive)
  """
  # if a calibration was not provided, use pixel number.  This is OK because the calibration is only used for plotting
  if (calibration == None):
    calibration = range(inframe.Data.shape[0])
  #if calibration size makes no sense, warn, and revert to pixel number
  if (len(calibration) != inframe.Data.shape[0] ):
    warnings.warn("calibration vector size doesn't match frame dimension, reverting to uncalibrated for optimalExtractWrapper",UserWarning)
    print "calibration size: ",len(calibration)," shape[0]:",inframe.Data.shape[0] 
    calibration = range(inframe.Data.shape[0])
  #get the profile
  spectra = []
  variances=[]
  skies=[]
  i=1
  if (interactive):
    #display a picker window for the user to select the integration limits for spectrum extraction
    while (raw_input("would you like to select a spectrum for extraction (yes or no): ").lower()[0] == 'y'):
      initialprofile=collapse2Dto1D(inframe.Data,0,mode='sum')
      initialsmooth=boxSmooth1D(initialprofile,9)
      picker = interactiveLimitPick(initialprofile,initialsmooth,frameSetup['spatialFitSpan'])
      peakBounds = picker.getSpectralRegion()
      if (len(peakBounds) != 2): continue
      ignored = picker.getIgnoredRegions()
      spec,var,sky = singleOptimal(inframe,peakBounds,ignored,frameSetup,name=name+"-"+str(i),calibration=calibration,telluric=telluric,fluxCal=fluxCal,tracefile=tracefile,savePath=savePath)
      spec[spec<0]=0
      spectra.append(spec)
      variances.append(var)
      skies.append(sky)
      i+=1
  else: #automatic peak and sky finding
    additionalPeaks=1 #just to make sure we don't get an infinite loop
    while(additionalPeaks):
      oneD=collapse2Dto1D(inframe.Data,0,mode='sum')
      smooth=boxSmooth1D(oneD,9)
      #subtract off a sky like amount and find peak/extent...
      fit = fit1D(oneD,2,frameSetup['spatialFitSpan'])
      fitted = poly2oneD(fit,len(oneD))
      minusFit = oneD-fitted
      smoothMfit=smooth-fitted
      maxX = np.argmax(minusFit[frameSetup['spatialFitSpan'][0]:frameSetup['spatialFitSpan'][1]])
      maxX+=frameSetup['spatialFitSpan'][0]
      maxV = minusFit[maxX]
      #put something here to break if maxV is not aboveSky enough
      print "peak found at spatial coordinate:",maxX
      if (oneD[maxX]<(1+aboveSky)*fitted[maxX]):
        print "breaking peak finding loop because no sufficient peaks found"
        print oneD[maxX]," is less than ",1+aboveSky," times ",fitted[maxX]
        additionalPeaks=0
        break
      #create a phony profile and subract off a low order fit, and a small constant
      #  times the value of the peak.  Then find where this fuction crosses zero.
      #  use these crossings as the min and max bounds of the peak.  This is broad, but that is 
      #  appropriote for Horne extraction
      smoothMoffset=smooth-fitted-maxV*.10*np.ones(len(oneD))
      zeroBounds=[seekZero(smoothMfit,maxX,-1),seekZero(smoothMfit,maxX,1)]
      width=zeroBounds[1]-zeroBounds[0]
      #use a sky window on each side of the peak, of 2 times the width of the peak starting .1* width of the peak removed from the peak
      #  this is tracked by regions that are ignored, not regions that are kept...  including the peak.
      ignored = [[0,int(round(max(frameSetup['spatialFitSpan'][0],zeroBounds[0]-2.1*width)))],
                 [int(round(zeroBounds[0]-.1*width)),int(round(zeroBounds[1]+.1*width))],
                 [int(round(min(frameSetup['spatialFitSpan'][1],zeroBounds[1]+2.1*width))),inframe.Data.shape[1]]]
      spec,var,sky = singleOptimal(inframe,zeroBounds,ignored,frameSetup,name=name+"-"+str(i),calibration=calibration,telluric=telluric,fluxCal=fluxCal,tracefile=tracefile,savePath=savePath)
      inframe=removeFromImage(inframe,zeroBounds)
      inframe.toFile(os.path.join(savePath,name+"-"+str(i)+"-depeaked"+USERDefs.extension))
      spec[spec<0]=0
      spectra.append(spec)
      variances.append(var)
      skies.append(sky)
      i+=1
      additionalPeaks-=1
  return spectra,variances,skies
      

def singleOptimal(inframe,bounds,ignored,frameSetup,name="frameNameNotSpecified",calibration=None,tracefile=None,telluric=None,fluxCal=None,savePath=USERDefs.savepath):
  """
  perform Horne extraction, and all that goes with it for the peak between bounds, usign the sky regions defined by ignored
  """

  # if a calibration was not provided, use pixel number.  This is OK because the calibration is only used for plotting
  if (calibration == None):
    calibration = range(inframe.Data.shape[0])
  #plot up, save, and maybe display the regions involved
  print "DEBUG: ignored regions in singleOptimal:",ignored
  plt.title(name)
  initialprofile=collapse2Dto1D(inframe.Data,0,mode='sum')  #only computed for plotting
  plt.plot(initialprofile,'b-')
  plt.axis((0,len(initialprofile),0,max(initialprofile[frameSetup['spatialFitSpan'][0]:frameSetup['spatialFitSpan'][1]])))
  plt.axvline(x=ignored[0][1],color='c')
  plt.axvline(x=ignored[-1][0],color='c')
  for i in range(1,len(ignored)-1):
    plt.axvline(x=ignored[i][0],color='c')
    plt.axvline(x=ignored[i][1],color='c')
  plt.axvline(x=bounds[0],color='m')
  plt.axvline(x=bounds[1],color='m')
  plt.xlabel("spatial")
  plt.savefig(os.path.join(savePath,name+"-peak_sky.png"))
  if (USERDefs.display):
    plt.show()
  plt.clf()

  #create sky images, and appropriotely trimmed images, etc...
  skyimage = makeSky(inframe,ignored)   #DEBUG
  skyimage.toFile(os.path.join(savePath,name+"-skyImage"+USERDefs.extension))
  #skyimage = makeSkyTraced(inframe,ignoredframeSetup)
  fixedImage = correctSkyTrace(inframe,ignored,int((bounds[0]+bounds[1])/2),frameSetup)
  fixedImage.toFile(os.path.join(savePath,name+"-skyTraceImage"+USERDefs.extension))
  skyimage = makeSky(fixedImage,ignored)
  skyimage.toFile(os.path.join(savePath,name+"-tracedSkyImage"+USERDefs.extension))
  spectrumSky=simpleExtract(skyimage.Data)
  subSky = frameSubtract(fixedImage,skyimage)
  subSky.toFile(os.path.join(savePath,name+"-skySubtracted"+USERDefs.extension))
  trimmed = copy.deepcopy(subSky)
  trimmed.Data = np.array(subSky.Data[:,bounds[0]:bounds[1]])
  trimmed.Data[np.isnan(trimmed.Data)]=0
  trimmed.toFile(os.path.join(savePath,name+"-trimmed"+USERDefs.extension))
  trimSky = copy.deepcopy(skyimage)
  trimSky.Data = np.array(skyimage.Data[:,bounds[0]:bounds[1]])
  trimSky.Data[np.isnan(trimSky.Data)]=0
  trimSky.toFile(os.path.join(savePath,name+"-trimmedSky"+USERDefs.extension))
  if (tracefile):
    tr = loadlist(tracefile)
  else:
    tr = trace(trimmed,fitspan=frameSetup['illuminatedSpan'])
    #print "DEBUG: Trace vector: ",tr
    savelist(tr,os.path.join(savePath,name+"_traceVector"+".dat"))
  rotated=correctTrace(subSky,tr)
  rotated.toFile(os.path.join(savePath,name+"-traceCorrected"+USERDefs.extension))
  trimRot = copy.deepcopy(rotated)
  trimRot.Data = np.array(rotated.Data[:,bounds[0]:bounds[1]])
  trimRot.Data[np.isnan(trimRot.Data)]=0
  trimRot.toFile(os.path.join(savePath,name+"-traceCorrected_trimmed"+USERDefs.extension))

  #get an initial spectrum to work with
  spectrum=np.sum(trimRot.Data,1)
  smoothSpectrumF=boxSmooth1D(spectrum)
  initialSpec = spectrum
  initialSmooth=smoothSpectrumF
  #print "DEBUG: inframe.Head:",inframe.Head
  Gain = inframe.Head['GAIN']
  RON  = inframe.Head['RON']
  variance=np.ones(trimRot.Data.shape)*RON**2
  #core of Horne algorithm.  Please see Horne 1986 for explaination.
  for i in range(USERDefs.maxExtractIter):
    trimmedData = trimRot.Data
    profileData = trimmedData/oneD2twoD(smoothSpectrumF,trimmedData.shape[1],1)    
    profileData[np.isnan(profileData)]=0
    profileData[profileData<0]=0
    sigmaP=np.sum(profileData,1)
    profileNorm=profileData/oneD2twoD(sigmaP,profileData.shape[1],1)
    newD=oneD2twoD(smoothSpectrumF,trimmedData.shape[1],1)*profileNorm+trimSky.Data
    variance=variance+newD/Gain
    prevspec = spectrum
    spectrum=np.sum(profileNorm*trimmedData/variance,1)/np.sum(profileNorm*profileNorm/variance,1)
    spectrum[np.isnan(spectrum)]=0
    smoothSpectrumF=boxSmooth1D(spectrum)
    if (np.max(spectrum-prevspec) < USERDefs.extractConvergenceVal):
      print "sufficiently converged at iteration "+str(i)
      break

  if (telluric is not None):
    spectrum=np.array(spectrum)*np.array(telluric)
    #TODO-track variance
  if (fluxCal is not None):
    spectrum=np.array(spectrum)*np.array(fluxCal)
    #TODO-track variance
  varianceFrm = copy.deepcopy(trimRot)
  varianceFrm.Data = variance
  varianceFrm.Head['history']= "variance of the extraction region"
  varianceFrm.toFile(os.path.join(savePath,name+"-varianceImage"+USERDefs.extension))
  spectrumV = np.sum(profileNorm,1)/np.sum(profileNorm*profileNorm/variance,1)
  #print "DEBUG: nan hunting:  profileNorm:",np.where(np.isnan(profileNorm))
  #print "DEBUG: nan hunting:  variance   :",np.where(np.isnan(variance))
  spectrumV[np.isnan(spectrumV)]=0
  savelist(spectrum,os.path.join(savePath,name+"_spectrum"+".dat"))
  savelist(spectrumV,os.path.join(savePath,name+"_variance"+".dat"))
  

  #make some plots and get out of here
  plt.title("Standard Spectrum of "+name)
  plt.ylabel("intensity?")
  plt.xlabel("wavelength")
  smoothSpectrumF[smoothSpectrumF<0]=0
  initialSmooth[initialSmooth<0]=0
  plt.plot(calibration,smoothSpectrumF,'b-')
  plt.plot(calibration,initialSmooth,'r-')
  plt.savefig(os.path.join(savePath,name+"-optimalSpectrum.png"))
  if (USERDefs.display):
    plt.show()
  plt.clf()
  plt.title("Spectrum Variance of "+name)
  plt.ylabel("Variance")
  plt.xlabel("wavelength")
  plt.plot(calibration,spectrumV,'b-')
  plt.savefig(os.path.join(savePath,name+"-varianceSpectrum.png"))
  if (USERDefs.display):
    plt.show()
  plt.clf()
  return spectrum,spectrumV,spectrumSky


def makeSky(inframe,ignored,order=1):
  """
  create a sky image
  """
  #create a sky image, ignoring the regions defined by the coordinate pairs in ignored
  skyFrm=copy.deepcopy(inframe)
  skyFrm.type="sky"
  skyFrm.Head['history']= "sky image of order "+str(order)
  #go row by row and fit a polynomial through the non-ignored part of the image, then rebuild an image from that polynomial.
  sky=[]
  ignored.reverse()
  #print "DEBUGGING:  ignored:",ignored 
  for i in range(inframe.Data.shape[0]):
    x=inframe.Data[i]
    y=range(0,x.size)
    for j in ignored:
      x=np.delete(x,range(j[0],j[1]))
      y=np.delete(y,range(j[0],j[1]))
    #have despectrummed sky row, fit it and fill it...
    #poly=fit1DNonuniform(x,y,order)
    poly=fit1DNonuniform(x,y,0)
    rebuilt = poly2oneD(poly,len(inframe.Data[i]))
    sky.append(rebuilt)
  skyFrm.Data=np.array(sky)
  return skyFrm

def correctSkyTrace(inframe,ignored,peakCoord,frameSetup,order=1):
  """
  use simple extraction to find peaks in the sky, then find the trace along a sky line, and correct the image such that the sky line is 
  perfectly horizontal.
  """
  skyClassic = makeSky(inframe,ignored,order)
  #skyClassic.toFile("products/classicSkyStepOfTracedSky.fits")  #DEBUG
  skySpectrum= simpleExtract(skyClassic.Data)
  smoothSpec = boxSmooth1D(skySpectrum)
  #at this point smoothSpec should be a smoothed spectrum of the sky lines only, as any objects should have been ignored
  if (USERDefs.debug and False):
    print "DEBUG: sky specturm: ",smoothSpec.tolist()
    print "DEBUG: restricted: ",smoothSpec[frameSetup['illuminatedSpan'][0]:frameSetup['illuminatedSpan'][1]].tolist()
    plt.title("Sky Spectrum")
    plt.ylabel("intensity?")
    plt.xlabel("spectral pixel coordinate")
    plt.plot(smoothSpec.tolist(),'b-')
    plt.ylim([0,max(smoothSpec.tolist())])
    plt.savefig("skySpectrum.png")
    plt.clf()
  #convert to use median of several sky lines traces
  skyPeaks=findPeaks(smoothSpec,spectralFitSpan=frameSetup['illuminatedSpan'])
  traces=[]
  for p in skyPeaks:
    if (USERDefs.debug):
      print "DEBUG: collecting trace at peak ",p
    p=int(round(p))
    justLine=copy.deepcopy(inframe)
    skyLine =np.array(inframe.Data[range(p-15,p+15),:])
    likeSpec = np.rot90(skyLine)
    justLine.Data = likeSpec
    #get the trace of the skyline
    tr = trace(justLine,order,frameSetup['spatialFitSpan'])
    if (USERDefs.debug):
      mxt= max(tr)
      mnt= min(tr)
      print "DEBUG: trace at ",p," goes from ",mnt," to ",mxt
    traces.append(tr)
  #take per column median
  traces = np.array(traces)
  if (USERDefs.debug):
    print "DEBUG: shape of traces array: ",traces.shape
  tr = np.median(traces,axis=0)




  #maxX = np.argmax(smoothSpec[frameSetup['illuminatedSpan'][0]:frameSetup['illuminatedSpan'][1]])
  #print "DEBUG: maxX pre addition: " , maxX, " span increases this by: ",frameSetup['illuminatedSpan'][0]
  #maxX+= frameSetup['illuminatedSpan'][0]
  #maxV = smoothSpec[maxX]
  #print "correcting sky trace based on line around pixel ",maxX," value there is ",maxV
  #justLine=copy.deepcopy(inframe)
  #skyLine =np.array(inframe.Data[range(maxX-15,maxX+15),:])
  #likeSpec = np.rot90(skyLine)
  #justLine.Data = likeSpec
  ##get the trace of the skyline
  #tr = trace(justLine,order,frameSetup['spatialFitSpan'])
  if (USERDefs.debug):
    print "DEBUG:  the sky line trace is : ",tr
  #rotate the input, correct the trace, and rotate the input back to it's starting point.  This will give a version of the imput in which the sky
  # lines are horizontal
  rotInput = copy.deepcopy(inframe)
  rotInput.Data = np.rot90(inframe.Data)
  #rotInput.toFile("products/rotatedInputStepOfTracedSky.fits")  #DEBUG
  rotTraced=correctTrace(rotInput,tr,peakCoord,direction=-1)
  #rotTraced.toFile("products/rotatedCorrectedStepOfTracedSky.fits")  #DEBUG
  outTraced=copy.deepcopy(inframe)
  outTraced.Data = np.rot90(rotTraced.Data,3)
  #outTraced.toFile("products/rotatedBackStepOfTracedSky.fits")  #DEBUG
  outTraced.Head['history']="Sky trace correction applied of order "+str(order)+"."
  return outTraced

  

def trace(inframe,order=3,fitspan=None):
  """
  trace the peak and return shift from vertical
  """
  if (fitspan is None):fitspan=[0,0]
  #if fitspan is not provided or is nonsense, use the whole chip
  if (fitspan[0] >= fitspan[1]):
    fitspan = [0,inframe.Data.shape[1]]
  #first smooth the peak so that pixel level variations accross columns are less likely to cause the peak to be inconsistant from row to row.
  #todo replace the smoothing and argmax with a polynomial fit, and peak finding so that I can find the max between pixels  for now integers will do
  smoothed = []
  for i in range(inframe.Data.shape[1]):
    smoothed.append(boxSmooth1D(inframe.Data[:,i],9))
  smoothed = np.array(smoothed)
  #find peak in each row
  maxX=np.argmax(smoothed,0)
  #fit a polynomial to eliminate pixel to pixel variation in the row direction
  #print "DEBUG: maxX:",maxX
  poly = fit1D(maxX,order,fitspan)
  tr = poly2oneD(poly,inframe.Data.shape[0])
  mid=np.median(tr[fitspan]).astype(float)
  tr = tr-np.ones(len(tr))*mid
  #print "DEBUG: trace:",tr
  return tr

def correctTrace(inframe,trace,maintainZeroPoint=None,direction=1):
  """
  correct the trace of the frame to the object of concern is vertical
  """
  #print "trace:",trace
  #use the scipy interp shift fuction to realign the image so that the trace of the peak is vertical.  will need to work on a per row basis
  corrected = copy.deepcopy(inframe)
  corrected.Head['HISTORY']="trace realigned"
  for i in range(inframe.Data.shape[0]):
    corrected.Data[i]=scipy.ndimage.shift(inframe.Data[i],direction*trace[i],mode='wrap',order=1)
  if (maintainZeroPoint is not None):
    #shift the data block by the oposite of the trace value at zeroPoint coordinate, main use is to maintain calibration accuracy
    print "DEBUG:  shifting image based on trace of ",trace[maintainZeroPoint]," at pixel coordinate ",maintainZeroPoint,"."
    corrected.Data=scipy.ndimage.shift(corrected.Data,[0,-trace[maintainZeroPoint]],mode='wrap',order=1)
  return corrected
  

def removeFromImage(inframe,bounds):
  """
  replace a region of the frame with a linear interpolation across it.
  """
  subtracted = copy.deepcopy(inframe)
  for i in range(inframe.Data.shape[0]):
    subtracted.Data[i,bounds[0]:bounds[1]] = np.linspace(inframe.Data[i,bounds[0]],inframe.Data[i,bounds[1]],bounds[1]-bounds[0])
  subtracted.Head['HISTORY']="peak replaced with linear interpolation across the space it occupied"
  return subtracted

    
def response(inframe,order=6,fitSpan=None,countThreshold=0):
  """
  generate a response fuction frame
  """
  if (fitSpan is None):fitSpan=[0,0]
  span = inframe.Data.shape[1]
  #print "DEBUG:input shape for response: ",inframe.Data.shape
  offset=int(math.floor(span*.1))  #drop 10% from each side so you only get the good part
  window=(offset,span-offset)
  body = trimArray(inframe.Data,window,1)
  oneDRaw=collapse2Dto1D(body,1)
  oneD=oneDRaw
  #oneD = boxSmooth1D(oneDRaw,int(body.shape[0]/100))
  #print "DEBUG:our oneD array is of length: ",len(oneD)
  #print oneD
  if (countThreshold and fitSpan[1]==0):
    #stop when counts drops below threshold
    initial=0
    for i in range(initial,len(oneD)):
      if (oneD[i]>countThreshold*1.1):
        initial = i
        print "response fit will start at ",initial
        fitSpan[0] = initial
        break
    final=initial+order+1
    for i in range(initial,len(oneD)):
      final = i
      fitSpan[1] = final
      if (oneD[i]<countThreshold/1.1):
        print "response fit will halt at ",final
        break
    if (final <= initial):
      final=body.shape[0]
      fitSpan[1] = final
  print "will fit on rows ",fitSpan
  fitParms = fit1D(oneD,order,fitSpan)
  recon1D=poly2oneD(fitParms,body.shape[0])
  #print "DEBUG: is reconstructed 1D too small?  ",len(recon1D)
  twoD = oneD2twoD(recon1D,inframe.Data.shape[1],1)
  #create the frame object to hold this
  respFrm=copy.deepcopy(inframe)
  respFrm.Data=twoD
  respFrm.type="Response"
  respFrm.Head['history']= "response order "+str(order)+" on pixels "+str(fitSpan)
  return respFrm
  

def oneD2twoD(oneD,size,axis=1):
  """
  copy a vector perpendicular to it's axis to fill a 2D array
  """
  list = [oneD for i in range(size)]
  arr = np.array(list)
  if axis==1:
    return arr.transpose()
  elif axis==0:
    return arr
  else:
    raise Exception("illegal axis specified.  Please specify 0 or 1")
  return


def fit1D(oneDarray,order=3,fitSpan=None,weights=None):
  """
  fit a polynomial to equally spaced points
  """
  if (fitSpan is None):
    fitSpan=[0,oneDarray.size]
    #fitSpan[0]=0
    #fitSpan[1]=oneDarray.size
  if (weights==None):
    weights=np.ones(oneDarray.size)
  #x=range(0,fitSpan[1]-fitSpan[0])    #this doesn't seem right to me, need to look at if x should be starting at 0.
  x=range(fitSpan[0],fitSpan[1]) 
  y=oneDarray[fitSpan[0]:fitSpan[1]]
  weight=weights[fitSpan[0]:fitSpan[1]]
  #print "fit span:",fitSpan
  #print "oneDarray size:",oneDarray.shape
  #print "x: "+str(len(x))+" y: "+str(len(y))
  p=np.polyfit(x,y,order,w=weight).tolist()
  p.reverse()
  return p

def fit1DNonuniform(oneDarray,indices,order=3,weights=None):
  """
  fit a polynomial to nonequally spaced points
  """
  #print "DEBUG:  investigating inputs to polyfit in fit1Dnonuniform: arguments in order"
  #print "arg1:  ",indices
  #print "arg2:  ",oneDarray
  #print "arg3:  ",order
  oneDarray=np.array(oneDarray)
  if (weights==None):
    weights=np.ones(oneDarray.size)
  p=np.polyfit(indices,oneDarray,order,w=weights).tolist()
  p.reverse()
  return p

def poly2oneD(poly,size):
  """
  return a vector representation of a polynomial
  """
  return np.polynomial.polynomial.polyval(range(0,size),poly)

def poly2oneDNonuniform(poly,grid):
  """
  return a vector representation of a polynomial
  """
  return np.polynomial.polynomial.polyval(grid,poly)


def boxSmooth1D(inarr,window=9):
  """
  box car smoothing
  """
  smoothed=np.convolve(inarr,np.ones(window)/window)
  smoothed = np.delete(smoothed,range(int(np.floor(window/2))),None)
  smoothed = np.delete(smoothed,range(len(smoothed)-int(np.floor(window/2)),len(smoothed)),None)
  #print "smoothdebug: "+str(len(inarr))+"  "+str(len(smoothed))
  return smoothed
    
def trimArray(data,window,axis):
  """
  remove outer part of the array
  """
  if (axis==1):
    return np.array(data[:,window[0]:window[1]])
  elif (axis==0):
    return np.array(data[window[0]:window[1],:])
  else:
    raise Exception("illegal axis specified in trimArray.  Please specify 0(spatial) or 1(spectral)")

def collapse2Dto1D(data,compressionAxis=1,mode='median'):
  """
  collapse a 2D field on to 1D
  """
  if mode=='median':
    return np.median(data,axis=compressionAxis)
  elif mode=='sum':
    return np.sum(data,axis=compressionAxis)
  else:
    raise RuntimeError('illegal mode in SpecReduceUtils.collapse2Dto1D')

def badPix(body,BPvalue=65000):
  """
  replace bad pixels with the average of the surounding pixels, (iteratively in the case of bad pixels bounded by bad pixels)
  """
  badpixcount = np.sum(np.where(body >= BPvalue,1,0))
  #as fraction of total?
  pixelcount = np.sum(np.where(body != -1000000000000.314159 ,1,0))     #I want to use None,
  #                                                                 or some other kind of undefined value here, 
  #                                                                rather than -big number, but it forces where 
  #                                                                to look at the array as a whole, rather than element by element.
  #                                                                   ultimately this would be likely to not have an impact, 
  #                                                                and if there was an impact it would only be a false warning.  
  #                                                                I will leave this for now and move on.
  badFraction=badpixcount/pixelcount
  print "starting with "+str(badpixcount)+" bad pixels, out of "+str(pixelcount)+" total pixels."
  if (badFraction >=USERDefs.BadPixelWarningFraction):
    warnings.warn("Bad Pixel Count is High",UserWarning)
  maxiter=20
  retArray = np.copy(body)
  iter=0
  while (badpixcount>0 and iter < maxiter):
    newret=np.copy(retArray)
    BParr=np.where(retArray >= BPvalue,1,0)
    BPcoords=np.nonzero(BParr)
    print "DEBUG  BPcoords: ",BPcoords
    print "DEBUG  BPcoords[0]: ",BPcoords[0]
    print "DEBUG  len(BPcoords[0]): ",len(BPcoords[0])
    #for B in range(BPcoords[0].size()):
    for B in range(len(BPcoords[0])):
      sum=0
      count=0
      for dir in ((-1,0),(0,-1),(1,0),(0,1)):
        print "DEBUG  anything a tuple? ",B,"   dir:",dir[0]
        print "DEBUG: something is a tuple..."
        print "DEBUG 1 entry:  ",BPcoords[0][B],"  next:",dir[0]
        print "DEBUG 2 entry:  ",BPcoords[1][B],"  next:",dir[0]
        print "DEBUG 3 entry:  ",BPcoords[0][B],"  next:",dir[1]
        print "DEBUG 4 entry:  ",BPcoords[0][B],"  next:",dir[1]
        if ((0<=(BPcoords[0][B]+dir[0])<=body.shape[0]) and (0<=(BPcoords[1][B]+dir[1])<=body.shape[1]) and not BParr[(BPcoords[0][B]+dir[0])][(BPcoords[1][B]+dir[1])]):
          count+=1
          sum+=body[(BPcoords[0][B]+dir[0]),(BPcoords[1][B]+dir[1])]
      if (count):
        newret[BPcoords[0][B],BPcoords[1][B]]=sum/count
    retArray=newret
    iter+=1
    badpixcount = np.sum(np.where(retArray >= BPvalue,1,0))
    print str(badpixcount)+" bad pixels after "+str(iter)+" iterations."
  return retArray

def spectrumBlend(WLcal1,Vals1,WLcal2,Vals2):
  """
  blend the overlap region and return a combined spectrum with combined calibration
  """
  regionStop=max(WLcal1)
  regionStart=min(WLcal2)
  
  gridSize=max(sum(1 for i in WLcal1 if i>regionStart),sum(1 for i in WLcal2 if i<regionStop))
  mixedCal = np.linspace(regionStart,regionStop,gridSize)

  Cal1Trunc = [WLcal1[i] for i in range(len(WLcal1)) if WLcal1[i] < regionStart]
  Val1Trunc = [Vals1[i] for i in range(len(WLcal1)) if WLcal1[i] < regionStart]
  Cal2Trunc = [WLcal2[i] for i in range(len(WLcal2)) if WLcal2[i] > regionStop]
  Val2Trunc = [Vals2[i] for i in range(len(WLcal2)) if WLcal2[i] > regionStop]

  MixedVal1 = np.interp(mixedCal,WLcal1,Vals1)
  MixedVal2 = np.interp(mixedCal,WLcal2,Vals2)
  mixFraction = np.linspace(0,1,gridSize)
  mixedVals = (1-mixFraction)*MixedVal1+mixFraction*MixedVal2

  WLCal = np.concatenate((Cal1Trunc,mixedCal,Cal2Trunc))
  Vals  = np.concatenate((Val1Trunc,mixedVals,Val2Trunc))

  return WLCal,Vals

def sigmaclip(rawCal,grid,calVar,frameSetup,clipFactor=3,order=3):
  """
  perform sigma clipping on a spectrum to remove fluctuations
  """
  maxiter=10
  count=1
  itercount=1
  loopCal = [i for i in range(len(rawCal)) if (i<=frameSetup['illuminatedSpan'][1] and i>=frameSetup['illuminatedSpan'][0])]
  mask = range(frameSetup['illuminatedSpan'][0],frameSetup['illuminatedSpan'][1]+1)
  maskedGrid = grid[mask[:]]
  print "DEBUG:  len(mask): ",len(mask)," len(maskedcal) ",len(loopCal)
  while (count):
    #print "DEBUG: iteration: ",itercount," raw: ",", ".join('%f'%x for x in loopCal)
    #print "DEBUG: iteration: ",itercount," var: ",", ".join('%f'%x for x in calVar)
    fitParms = fit1DNonuniform(loopCal,maskedGrid,order)
    #recon = poly2oneD(fitParms,len(rawCal))
    recon = poly2oneDNonuniform(fitParms,grid)
    #print "DEBUG: iteration: ",itercount," recon: ",", ".join('%f'%x for x in recon)
    count = sum([abs(rawCal[i]-recon[i])>clipFactor*calVar[i]**.5 for i in range(len(rawCal))])
    print "DEBUG: at iteration ",itercount," the count of devient entries is ",count," the fit parameters are ", fitParms
    if (itercount>=maxiter):
      return recon
    itercount+=1
    if (count >0):
      loopCal=recon
      mask = range(len(rawCal))
      keepers = [abs(rawCal[i]-recon[i])<=clipFactor*calVar[i] for i in range(len(rawCal))]
      loopCal = [loopCal[i] for i in range(len(loopCal)) if (keepers[i] and i<=frameSetup['illuminatedSpan'][1] and i>=frameSetup['illuminatedSpan'][0])]
      mask = [mask[i] for i in range(len(mask)) if (keepers[i] and i<=frameSetup['illuminatedSpan'][1] and i>=frameSetup['illuminatedSpan'][0])]
      maskedGrid = grid[mask[:]]
  return recon

def computeTelluricOld(telluricSpec,calibration,ignores,order=3):
  """
  compute a telluric correction from the supplied spectrum
  """
  #remove ignored regions
  cal = np.array(calibration)
  tell= np.array(telluricSpec)
  while (ignores):
    region = ignores.pop()
    remove = np.where((calibration>region[0])*(calibration<region[1]))
    cal=np.delete(cal,remove)
    tell=np.delete(tell,remove)
  fitParms = fit1DNonuniform(tell,cal,order)
  #could sigmaclip in here, is it worth it?
  poly = poly2oneDNonuniform(fitParms,calibration)
  corr=np.array(poly)/np.array(telluricSpec)
  return corr  

def computeTelluric(telluricSpec,calibration,ignores,variance=None,order=3):  #assume variance due to telluric is insignificant
  """
  compute a telluric correction from the supplied spectrum
  """
  if (len(calibration) != len(telluricSpec)):
    warnings.warn("WARNING cannot compute a telluric correction with mismatched callibrations and spectra",UserWarning)
    print "DEBUG: telluric inputs: tellspec:",telluricSpec,"    calibration:",calibration,"     ignores:",ignores
  cal = np.array(calibration)
  tell= np.array(telluricSpec)
  corr = np.ones(len(telluricSpec))
  if (variance==None):
    variance = np.ones(len(telluricSpec))
  weight= 1.0/variance
  #print "DEBUG:  size check  var: ",variance.size,"   weight:",weight.size
  #corr = cal/cal
  #print "DEBUG: telluricSpec  ",telluricSpec
  for region in ignores:
    if (max(cal)<region[0] or min(cal)>region[1]):
      warnings.warn("WARNING attempting to apply a telluric correction to a region in which telluric doesn't apply",UserWarning)
      continue
    size = region[1]-region[0]
    #print "DEBUG: region: ",region[0],"  ",region[1]
    #fitspan = np.where((calibration>region[0]-size)*(calibration<region[1]+size))
    #tellspan= np.where((calibration>region[0])*(calibration<region[1]))
    fitspan = np.where((cal>region[0]-size)*(cal<region[1]+size))
    tellspan= np.where((cal>region[0])*(cal<region[1]))
    localCal= cal[fitspan]
    #print "DEBUG:region: ",region," fitspan: ",fitspan," localCal: ",localCal," tellspan: ",tellspan
    localVal= tell[fitspan]
    localW=weight[fitspan]
    localSpan=np.where((localCal>region[0])*(localCal<region[1]))
    localCal=np.delete(localCal,localSpan)
    localVal=np.delete(localVal,localSpan)
    localW=np.delete(localW,localSpan)
    #tellCal = calibration[tellspan]
    tellCal = cal[tellspan]
    tellVal = tell[tellspan]
    #print "DEBUG: localVal: ",localVal, " localCal: ",localCal, "  order:",order
    fitParms=fit1DNonuniform(localVal,localCal,order,weights=localW)
    regionRefit = poly2oneDNonuniform(fitParms,tellCal)
    regionCorr=tellVal/regionRefit
    corr[tellspan]=regionCorr
  return corr

def mergeSpecs(spec1,cal1,spec2,cal2):
  #combine them.  Get a new grid in the overlap area, interpolate both spectra onto it, then linspace from from one to the other

  regionStop =max(cal1)
  regionStart=min(cal2)
  
  #get the size of the grid we'll interpolate to
  gridSize=max(sum(1 for i in cal1 if i>regionStart),sum(1 for i in cal2 if i<regionStop))
  mixedCal = np.linspace(regionStart,regionStop,gridSize)
  Cal1Trunc = [cal1[i] for i in range(len(cal1)) if cal1[i] < regionStart]
  Spec1Trunc= [spec1[i] for i in range(len(cal1)) if cal1[i] < regionStart]
  Cal2Trunc  = [cal2[i] for i in range(len(cal2)) if cal2[i] > regionStop]
  Spec2Trunc = [spec2[i] for i in range(len(cal2)) if cal2[i] > regionStop]
  
  MixedVals1= np.interp(mixedCal,cal1,spec1)
  MixedVals2 = np.interp(mixedCal,cal2,spec2)
  
  mixFraction = np.linspace(0,1,gridSize)
  mixedSpec = (1-mixFraction)*MixedVals1+mixFraction*MixedVals2
  
  WLCal = np.concatenate((Cal1Trunc,mixedCal,Cal2Trunc))
  Spec  = np.concatenate((Spec1Trunc,mixedSpec,Spec2Trunc))
  return Spec,WLCal
  
  
def computeFluxCal(spectrum,calibration,ref):
  Fgrid = [row[0] for row in ref]
  FVal = [row[1] for row in ref]
  Flux = np.interp(calibration,Fgrid,FVal)
  return Flux

def loadFluxStandard(filename,columns=None):
  """
  open a flux standard reference file and return the wavelenght,flux
  """
  if (columns is None):columns=[0,2]
  with open(filename,'r') as f:
    linelist = [[float(line.rstrip('\n').split()[i]) for i in columns] for line in f]
  return linelist

def weightedMean(valueVariPairs):
  """
  compute an inverse variance weighted mean, and the new variance
  """
  #print "DEBUG weightedMean:  input: ",valueVariPairs
  invVaris=[1/p[1] for p in valueVariPairs]
  #print "DEBUG weightedMean:  invVaris: ",invVaris   
  variWeighted = [p[0]*p[1] for p in valueVariPairs]
  #print "DEBUG weightedMean:  weighted: ",variWeighted
  weightSum = reduce(lambda x, y: x+y,variWeighted)
  #print "DEBUG weightedMean:  weightSum: ",weightSum      
  variSum = reduce(lambda x,y:x+y,invVaris)
  #print "DEBUG weightedMean:  variSum: ",variSum      
  mean = weightSum*variSum
  #print "DEBUG weightedMean:  mean: ", mean     
  newVari=1/variSum
  #print "DEBUG weightedMean:  newVari: ",  newVari    

  #replace nan with 0 in return values...
  meannans=np.isnan(mean)
  mean[meannans]=0.0
  varinans=np.isnan(newVari)
  newVari[varinans]=0.0
  return mean,newVari

def savelist(my_list,filename):
  """
  write a list to file
  """
  with open(filename, 'w') as f:
    for s in my_list:
      f.write("%s\n" % s)

def loadlist(filename):
  """
  load a list from file
  """
  with open(filename, 'r') as f:
    my_list = [float(line.rstrip('\n')) for line in f]
  return my_list

def array2Fits(array, name, head=None):
  """write a numpy array to a fits file with a genaric header, when working with a frame use child method"""
  fitsdata=fits.PrimaryHDU(array,header=head)
  fitsdata.writeto(name,clobber=True)

def splitPathExt(instr):
  """
  split a filename into it's path (if present), the base file name, and it's extension
  """
  pathFN,ext = os.path.splitext(instr)
  path,FN = os.path.split(pathFN)
  return path,FN,ext
  
def fromInput(filename):
  """
  load and parse an input file for either modification in the GUI or to drive an analysis in the pipeline
  """
  inputStruct={}
  with open(filename,'r') as f:
    inputStruct=json.load(f)
    #sanity check/fill in blanks
    #TODO
    #if ("bias" in self.InputStruct):
    #  for bias in self.InputStruct["bias"]:
  return inputStruct

def fromInput_old(filename):
  """
  load and parse an input file for either modification in the GUI or to drive an analysis in the pipeline
  """
  #rewrite this fuction to collapse the if elifs to a loop...
  inputStruct={}
  inputStruct['names']={}
  with open(filename,'r') as f:
    for line in f:
      l=line.strip()
      if (len(l) == 0):
        continue
      if (l[0] == '#'):
        continue
      splitLine=l.split(':')
      action=splitLine[0]
      name=splitLine[1]
      arguments=splitLine[2].split(';')
      #print "DEBUG: action: ",action," name: ",name," arguments: ",arguments
      if (action=="bias"):
        if (action not in inputStruct):
          inputStruct[action]={}
        if (name not in inputStruct[action]):
          inputStruct[action][name]={}
        for arg in arguments:
          argVal = arg.split('=')
          if (argVal[0]=="frame"):
            inputStruct[action][name][argVal[0]]=argVal[1]
          elif (argVal[0]=="fromfile"):
            inputStruct[action][name][argVal[0]]=argVal[1]
          else:
            raise ValueError('Invalid argument type for bias:'+argVal[0])
            continue
      elif (action=="flat"):
        if (action not in inputStruct):
          inputStruct[action]={}
        if (name not in inputStruct[action]):
          inputStruct[action][name]={}
        for arg in arguments:
          argVal = arg.split('=')
          if (argVal[0]=="frame"):
            inputStruct[action][name][argVal[0]]=argVal[1]
          elif (argVal[0]=="bias"):
            inputStruct[action][name][argVal[0]]=argVal[1]
          elif (argVal[0]=="fromfile"):
            inputStruct[action][name][argVal[0]]=argVal[1]
          else:
            raise ValueError('Invalid argument type for flat:'+argVal[0])
            continue
      elif (action=="telluric_correct"):
        if (action not in inputStruct):
          inputStruct[action]={}
        if (name not in inputStruct[action]):
          inputStruct[action][name]={}
        for arg in arguments:
          argVal = arg.split('=')
          if (argVal[0]=="spectrum"):
            inputStruct[action][name][argVal[0]]=argVal[1]
          elif (argVal[0]=="fromfile"):
            inputStruct[action][name][argVal[0]]=argVal[1]
          else:
            raise ValueError('Invalid argument type for telluric:'+argVal[0])
            continue
      elif (action=="WL_calibrate"):
        if (action not in inputStruct):
          inputStruct[action]={}
        if (name not in inputStruct[action]):
          inputStruct[action][name]={}
        for arg in arguments:
          argVal = arg.split('=')
          if (argVal[0]=="frame"):
            inputStruct[action][name][argVal[0]]=argVal[1]
          elif (argVal[0]=="bias"):
            inputStruct[action][name][argVal[0]]=argVal[1]
          elif (argVal[0]=="ref"):
            inputStruct[action][name][argVal[0]]=argVal[1]
          elif (argVal[0]=="flat"):
            inputStruct[action][name][argVal[0]]=argVal[1]
          #elif (argVal[0]=="pairFile"):
          #  inputStruct[action][name][argVal[0]]=argVal[1]
          #elif (argVal[0]=="calFile"):
          #  inputStruct[action][name][argVal[0]]=argVal[1]
          elif (argVal[0]=="fromfile"):
            inputStruct[action][name][argVal[0]]=argVal[1]
          else:
            raise ValueError('Invalid argument type for WL_calibrate:'+argVal[0])
            continue
      elif (action=="WL_combo"):
        if (action not in inputStruct):
          inputStruct[action]={}
        if (name not in inputStruct[action]):
          inputStruct[action][name]={}
        for arg in arguments:
          argVal = arg.split('=')
          if (argVal[0]=="calibrationsets"):
            inputStruct[action][name][argVal[0]]=argVal[1]
          else:
            raise ValueError('Invalid argument type for WL_combo:'+argVal[0])
            continue
      elif (action=="spectrum"):
        if (action not in inputStruct):
          inputStruct[action]={}
        if (name not in inputStruct[action]):
          inputStruct[action][name]={}
        for arg in arguments:
          argVal = arg.split('=')
          if (argVal[0]=="frame"):
            inputStruct[action][name][argVal[0]]=argVal[1]
          elif (argVal[0]=="bias"):
            inputStruct[action][name][argVal[0]]=argVal[1]
          elif (argVal[0]=="WL_cal"):
            inputStruct[action][name][argVal[0]]=argVal[1]
          elif (argVal[0]=="telluric"):
            inputStruct[action][name][argVal[0]]=argVal[1]
          elif (argVal[0]=="flat"):
            inputStruct[action][name][argVal[0]]=argVal[1]
          elif (argVal[0]=="flux_calibrator"):
            inputStruct[action][name][argVal[0]]=argVal[1]
          elif (argVal[0]=="fromfile"):
            inputStruct[action][name][argVal[0]]=argVal[1]
          elif (argVal[0]=="skyCalCorrect"):
            inputStruct[action][name][argVal[0]]=argVal[1]
          else:
            raise ValueError('Invalid argument type for spectrum:'+argVal[0])
            continue
        if ("flux_calibrator" not in inputStruct[action][name]):
          inputStruct[action][name]["flux_calibrator"]="*none"
        if ("skyCalCorrect" not in inputStruct[action][name]):
          inputStruct[action][name]["skyCalCorrect"]="0"
        if ("WL_cal" not in inputStruct[action][name]):
          inputStruct[action][name]["WL_cal"]="*none"
          inputStruct[action][name]["skyCalCorrect"]="0"
        if ("telluric" not in inputStruct[action][name]):
          inputStruct[action][name]["telluric"]="*none"
      elif (action=="flux_calibrate"):
        if (action not in inputStruct):
          inputStruct[action]={}
        if (name not in inputStruct[action]):
          inputStruct[action][name]={}
        for arg in arguments:
          argVal = arg.split('=')
          if (argVal[0]=="spectrum"):
            inputStruct[action][name][argVal[0]]=argVal[1]
          elif (argVal[0]=="reference"):
            inputStruct[action][name][argVal[0]]=argVal[1]
          elif (argVal[0]=="fromfile"):
            inputStruct[action][name][argVal[0]]=argVal[1]
          else:
            raise ValueError('Invalid argument type for flux_calibrate:'+argVal[0])
            continue
      elif (action=="mergeSpectra"):
        if (action not in inputStruct):
          inputStruct[action]={}
        if (name not in inputStruct[action]):
          inputStruct[action][name]={}
        for arg in arguments:
          argVal = arg.split('=')
          if (argVal[0]=="red"):
            inputStruct[action][name][argVal[0]]=argVal[1]
          elif (argVal[0]=="blue"):
            inputStruct[action][name][argVal[0]]=argVal[1]
          else:
            raise ValueError('Invalid argument type for mergeSpectra:'+argVal[0])
            continue
      else:
        raise ValueError('Invalid action type in input:'+action)
        continue
      if (name in inputStruct['names']):
        raise KeyError('name already in use: '+name)
      inputStruct['names'][name]=action
  return inputStruct
      

def toInput(filename,inputStruct):
  """
  write the input information for an analysis to an input file
  """
  #stringVer=json.dumps(inputStruct)
  with open(filename, 'w') as f:
    json.dump(inputStruct,f,indent=2)

def toInput_old(filename,inputStruct):
  """
  write the input information for an analysis to an input file
  """
  # bias flat wl_calibrate wl_combo spectrum flux_calibrate telluric_correct combineSpectra
  fields = ["bias","flat","wl_calibrate","wl_combo","spectrum","flux_calibrate","telluric_correct","combineSpectra"]
  with open(filename, 'w') as f:
    #print "DEBUG: inputStruct[names]:  ",inputStruct['names']
    #for name in inputStruct['names']:
    #  action = inputStruct['names'][name]
    #  print "DEBUG:  name:action   ",name,":",action
    for name in inputStruct['names']:
      action = inputStruct['names'][name]
      argstring=""
      for arg in inputStruct[action][name]:
        #print "DEBUG:  working on: ",action,name,arg
        val=inputStruct[action][name][arg]
        #print "DEBUG:  it's: ",val
        argstring=argstring+";"+arg+"="+val
      line = action+":"+name+":"+argstring[1:]
      f.write("%s\n" % line)   

def files2Input(filelist):
  """
  parse header info from a list of fits files and construct an analysis control file
  """
  #collect headers to start
  inputStruct={}
  inputStruct['names']={}
  heads={}
  bluefiles=[]
  redfiles=[]
  biasimgs=[]
  flatimgs=[]
  lampimgs=[] #cal
  spectrumimgs=[]  #object
  for f in filelist:
    #only process red and blue at this time
    rawFits = fits.open(f)
    heads[f]=rawFits[0].header
    #print "DEBUG: what does it look like? \n",f,"\n ", repr(heads[f])
    if ('FPA' in heads[f]):
      fpa=heads[f]['FPA']
      fpa=fpa.strip().lower()
      if (fpa=='dbsp_blue'):
        bluefiles.append(f)
      elif (fpa=='dbsp_red2'):
        redfiles.append(f)
      else:
        warnings.warn("WARNING FPA card for file "+str(f)+" is not recognized.  FPA:"+str(heads[f]['FPA']),UserWarning)
        continue
    else:
      warnings.warn("WARNING FPA card for file "+str(f)+" is absent.",UserWarning)
      continue
    #fn = splitPathExt(f)[1]
    #if (fn.startswith('blue')):
    #  bluefiles.append(f)
    #elif (fn.startswith('red')):
    #  redfiles.append(f)
    #else:
    #  warnings.warn("WARNING fits file "+f+" is not clearly red or blue.  ignoring.",UserWarning)
    #  next
    if ('IMGTYPE' in heads[f]):
      imgtype=heads[f]['IMGTYPE']
      imgtype = imgtype.strip().lower()
      if (imgtype == 'bias'):
        biasimgs.append(f)
      elif (imgtype == 'flat'):
        flatimgs.append(f)
      elif (imgtype == 'cal'):
        lampimgs.append(f)
      elif (imgtype == 'object'):
        spectrumimgs.append(f)
      else:
        warnings.warn("WARNING image type card for file "+str(f)+" is not recognized.  IMGTYPE:"+str(heads[f]['IMGTYPE']),UserWarning)
    else:
      warnings.warn("WARNING image type card for file "+str(f)+" is absent.",UserWarning)
  #bias
  bblist=[]
  rblist=[]
  if (len(biasimgs) >0):
    inputStruct['bias']={}
  for bf in biasimgs:
    if bf in bluefiles:
      bblist.append(bf)
    elif bf in redfiles:
      rblist.append(bf)
    else:
      print "DEBUG:  this should never happen.  file is not red or blue"
      continue
  bbias=""
  if (len(bblist) > 0):
    inputStruct['names']["blue_bias"]='bias'
    inputStruct['bias']["blue_bias"]={}
    inputStruct['bias']["blue_bias"]['frame']=bblist
    inputStruct['bias']["blue_bias"]['fpa']='dbsp_blue'
    bbias="blue_bias"
  else:
    bbias="*over"
  rbias=""
  if (len(rblist) > 0):
    inputStruct['names']["red_bias"]='bias'
    inputStruct['bias']["red_bias"]={}
    inputStruct['bias']["red_bias"]['frame']=rblist
    inputStruct['bias']["red_bias"]['fpa']='dbsp_red2'
    rbias="red_bias"
  else:
    rbias="*over"
  #flat
  bflist=[]
  rflist=[]
  if (len(flatimgs) >0):
    inputStruct['flat']={}
  for ff in flatimgs:
    if ff in bluefiles:
      bflist.append(ff)
    elif ff in redfiles:
      rflist.append(ff)
    else:
      print "DEBUG:  this should never happen.  file is not red or blue"
      continue
  bflat=""
  if (len(bflist) >0):
    inputStruct['names']["blue_flat"]='flat'
    inputStruct['flat']["blue_flat"]={}
    inputStruct['flat']["blue_flat"]['frame']=bflist
    #inputStruct['flat']["blue_flat"]['frame']=fileListBuilder(bflist)
    inputStruct['flat']["blue_flat"]['bias']=bbias
    inputStruct['flat']["blue_flat"]['fpa']='dbsp_blue'
    bflat="blue_flat"
  else:
    bflat="*none"
  rflat=""
  if (len(rflist) >0):
    inputStruct['names']["red_flat"]='flat'
    inputStruct['flat']["red_flat"]={}
    inputStruct['flat']["red_flat"]['frame']=rflist
    #inputStruct['flat']["red_flat"]['frame']=fileListBuilder(rflist)
    inputStruct['flat']["red_flat"]['bias']=rbias
    inputStruct['flat']["red_flat"]['fpa']='dbsp_red2'
    rflat="red_flat"
  else:
    rflat="*none"
  bclist=[]
  rclist=[]
  if (len(lampimgs) >0):
    inputStruct['WL_calibrate']={}
  for lc in lampimgs:
    if lc in bluefiles:
      bclist.append(lc)
    elif lc in redfiles:
      rclist.append(lc)
    else:
      print "DEBUG:  this should never happen.  file is not red or blue"
      continue
  if (len(bclist)>0):
    #create a calibration entry for each lamp setting.  if there is more than one lamp setting create a combo and use that, else use the solitary one...
    lampdict={}
    bcal = "blue_cal"
    for f in bclist:
      if (heads[f]['LAMPS'].strip() == ''):
        warnings.warn("Warning:  Calibration lamp specification string is empty for file "+f +".  Not using this calibrator.",UserWarning)
        continue
      if (heads[f]['LAMPS'].strip() == '0000000'):
        warnings.warn("Warning:  Calibration lamp specification is no lamps for file "+f +".  Not using this calibrator.",UserWarning)
        continue
      if (heads[f]['LAMPS'] not in lampdict):
        lampdict[heads[f]['LAMPS']]=[]
      lampdict[heads[f]['LAMPS']].append(f)
    if (len(lampdict)==1):
      lampstr=lampdict.keys()[0]
      refs=[]
      for i in range(0,len(lampstr)):
        char=lampstr[i]
        try:
          useIt=int(char)
        except:
          warnings.warn("Warning:  Calibration lamp specification string "+lampstr+" contains a non-boolean character at position "+str(i)+".  Ignoring this lamp.",UserWarning)
          useIt=False
        if useIt:
          print "DEBUG: adding lamp bit ",i," from string ",lampstr, " file will be ",USERDefs.DBSPConstants['lampBits'][i]," in ",USERDefs.DBSPConstants['referenceDir']
          refs.append(os.path.join(USERDefs.DBSPConstants['referenceDir'],USERDefs.DBSPConstants['lampBits'][i]))
      inputStruct['names']["blue_cal"]='WL_calibrate'
      inputStruct['WL_calibrate']["blue_cal"]={}
      inputStruct['WL_calibrate']["blue_cal"]['frame']=lampdict[lampstr]
      inputStruct['WL_calibrate']["blue_cal"]['bias']=bbias
      inputStruct['WL_calibrate']["blue_cal"]['flat']="*none"
      inputStruct['WL_calibrate']["blue_cal"]['ref']=refs
      inputStruct['WL_calibrate']["blue_cal"]['fpa']='dbsp_blue'
    elif (len(lampdict) >1):
      if ('WL_combo' not in inputStruct):
        inputStruct['WL_combo']={}
      for lamp in lampdict:
        refs=[]
        for i in range(0,len(lamp)):
          char=lamp[i]
          try:
            useIt=int(char)
          except:
            warnings.warn("Warning:  Calibration lamp specification string "+lamp+" contains a non-boolean character at position "+str(i)+".  Ignoring this lamp.",UserWarning)
            useIt=False
          if useIt:
            print "DEBUG: adding lamp bit ",i," from string ",lamp, " file will be ",USERDefs.DBSPConstants['lampBits'][i]," in ",USERDefs.DBSPConstants['referenceDir']
            refs.append(os.path.join(USERDefs.DBSPConstants['referenceDir'],USERDefs.DBSPConstants['lampBits'][i]))
        inputStruct['names']["blue_cal_"+lamp]='WL_calibrate'
        inputStruct['WL_calibrate']["blue_cal_"+lamp]={}
        inputStruct['WL_calibrate']["blue_cal_"+lamp]['frame']=lampdict[lamp]
        inputStruct['WL_calibrate']["blue_cal_"+lamp]['bias']=rbias
        inputStruct['WL_calibrate']["blue_cal_"+lamp]['flat']="*none"
        inputStruct['WL_calibrate']["blue_cal_"+lamp]['ref']=refs
        inputStruct['WL_calibrate']["blue_cal_"+lamp]['fpa']='dbsp_blue'
      inputStruct['names']["blue_cal"]='WL_combo'
      inputStruct['WL_combo']["blue_cal"]={}
      inputStruct['WL_combo']["blue_cal"]['calibrationsets']=["blue_cal_"+s for s in lampdict.keys()]
    else:
      print "DEBUG:  Lampdict blue should not be empty here"
      bcal = "*none"
  else:
    bcal = "*none"
  if (len(rclist)>0):
    #create a calibration entry for each lamp setting.  if there is more than one lamp setting create a combo and use that, else use the solitary one...
    lampdict={}
    rcal = "red_cal"
    for f in rclist:
      #print "DEBUG:  ",f," is a lamp file with lampstr ",heads[f]['LAMPS']
      if (not heads[f]['LAMPS'].strip() or heads[f]['LAMPS'].strip()==' '):
        warnings.warn("Warning:  Calibration lamp specification string is empty for file "+f +".  Not using this calibrator.",UserWarning)
        continue
      if (heads[f]['LAMPS'].strip() == '0000000'):
        warnings.warn("Warning:  Calibration lamp specification is no lamps for file "+f +".  Not using this calibrator.",UserWarning)
        continue
      if (heads[f]['LAMPS'] not in lampdict):
        #print "DEBUG: adding ",heads[f]['LAMPS']," to lampdict from file ",f
        lampdict[heads[f]['LAMPS']]=[]
      lampdict[heads[f]['LAMPS']].append(f)
    if (len(lampdict)==1):
      lampstr=lampdict.keys()[0]
      refs=[]
      for i in range(0,len(lampstr)):
        char=lampstr[i]
        try:
          useIt=int(char)
        except:
          warnings.warn("Warning:  Calibration lamp specification string "+lampstr+" contains a non-boolean character at position "+str(i)+".  Ignoring this lamp.",UserWarning)
          useIt=False
        if useIt:
          print "DEBUG: adding lamp bit ",i," from string ",lampstr, " file will be ",USERDefs.DBSPConstants['lampBits'][i]," in ",USERDefs.DBSPConstants['referenceDir']
          refs.append(os.path.join(USERDefs.DBSPConstants['referenceDir'],USERDefs.DBSPConstants['lampBits'][i]))
      inputStruct['names']["red_cal"]='WL_calibrate'
      inputStruct['WL_calibrate']["red_cal"]={}
      inputStruct['WL_calibrate']["red_cal"]['frame']=lampdict[lampstr]
      inputStruct['WL_calibrate']["red_cal"]['bias']=bbias
      inputStruct['WL_calibrate']["red_cal"]['flat']="*none"
      inputStruct['WL_calibrate']["red_cal"]['ref']=refs
      inputStruct['WL_calibrate']["red_cal"]['fpa']='dbsp_red2'
    elif (len(lampdict)>1):
      if ('WL_combo' not in inputStruct):
        inputStruct['WL_combo']={}
      for lamp in lampdict:
        refs=[]
        for i in range(0,len(lamp)):
          char=lamp[i]
          try:
            useIt=int(char)
          except:
            warnings.warn("Warning:  Calibration lamp specification string "+lamp+" contains a non-boolean character at position "+str(i)+".  Ignoring this lamp.",UserWarning)
            useIt=False
          if useIt:
            print "DEBUG: adding lamp bit ",i," from string ",lamp, " file will be ",USERDefs.DBSPConstants['lampBits'][i]," in ",USERDefs.DBSPConstants['referenceDir']
            refs.append(os.path.join(USERDefs.DBSPConstants['referenceDir'],USERDefs.DBSPConstants['lampBits'][i]))
        inputStruct['names']["red_cal_"+lamp]='WL_calibrate'
        inputStruct['WL_calibrate']["red_cal_"+lamp]={}
        inputStruct['WL_calibrate']["red_cal_"+lamp]['frame']=lampdict[lamp]
        inputStruct['WL_calibrate']["red_cal_"+lamp]['bias']=rbias
        inputStruct['WL_calibrate']["red_cal_"+lamp]['flat']="*none"
        inputStruct['WL_calibrate']["red_cal_"+lamp]['ref']=refs
        inputStruct['WL_calibrate']["red_cal_"+lamp]['fpa']='dbsp_red2'
      inputStruct['names']["red_cal"]='WL_combo'
      inputStruct['WL_combo']["red_cal"]={}
      inputStruct['WL_combo']["red_cal"]['calibrationsets']=["red_cal_"+s for s in lampdict.keys()]
    else:
      print "DEBUG:  Lampdict red should not be empty here"
      rcal = "*none"
  else:
    rcal = "*none"
  if (len(spectrumimgs) >0):
    inputStruct['spectrum']={}
  bilist=[]
  rilist=[]
  for si in spectrumimgs:
    if si in bluefiles:
      bilist.append(si)
    elif si in redfiles:
      rilist.append(si)
    else:
      print "DEBUG:  this should never happen.  file is not red or blue"
      continue
  if (len(bilist) >0):
    objdict={}
    for f in bilist:
      if (heads[f]['OBJECT'] not in objdict):
        objdict[heads[f]['OBJECT']] = []
      objdict[heads[f]['OBJECT']].append(f)
    for obj in objdict:
      inputStruct['names']['blue_'+obj]='spectrum'
      inputStruct['spectrum']['blue_'+obj]={}
      inputStruct['spectrum']['blue_'+obj]['frame']=objdict[obj]
      inputStruct['spectrum']['blue_'+obj]['bias']=bbias
      inputStruct['spectrum']['blue_'+obj]['flat']=bflat
      inputStruct['spectrum']['blue_'+obj]['WL_cal']=bcal
      inputStruct['spectrum']['blue_'+obj]['flux_calibrator']="*none"
      inputStruct['spectrum']['blue_'+obj]['telluric']="*none"
      inputStruct['spectrum']['blue_'+obj]['skyCalCorrect']=0
      inputStruct['spectrum']['blue_'+obj]['WL_cal']=bcal
      inputStruct['spectrum']["blue_"+obj]['fpa']='dbsp_blue'
  if (len(rilist) >0):
    objdict={}
    for f in rilist:
      if (heads[f]['OBJECT'] not in objdict):
        objdict[heads[f]['OBJECT']] = []
      objdict[heads[f]['OBJECT']].append(f)
    for obj in objdict:
      inputStruct['names']['red_'+obj]='spectrum'
      inputStruct['spectrum']['red_'+obj]={}
      inputStruct['spectrum']['red_'+obj]['frame']=objdict[obj]
      inputStruct['spectrum']['red_'+obj]['bias']=rbias
      inputStruct['spectrum']['red_'+obj]['flat']=rflat
      inputStruct['spectrum']['red_'+obj]['flux_calibrator']="*none"
      inputStruct['spectrum']['red_'+obj]['telluric']="*none"
      inputStruct['spectrum']['red_'+obj]['skyCalCorrect']=0
      inputStruct['spectrum']['red_'+obj]['WL_cal']=rcal
      inputStruct['spectrum']["red_"+obj]['fpa']='dbsp_red2'
  
  
  
  return inputStruct
  
  
  


     
  


def fileListParser(listString):
  """
  split a bracketed, quoted, comma delimeted string representation of a path list into a list of paths
  """
  justlist=listString.strip("[]")
  quotelist=justlist.split(",")
  bare = [x.strip("'") for x in quotelist]
  if (bare == [""]):
    return []
  return bare


def fileListBuilder(fileList):
  """
  create a bracketed, quoted,comma delimeted string repr of a list of paths
  """
  quot= ["'"+path+"'" for path in fileList]
  joined = ",".join(quot)
  return "["+joined+"]"


