#!/usr/bin/env python


import Tkinter as tk
import tkFileDialog
import tkMessageBox
import tkSimpleDialog
import os
from astropy.io import fits
import time #debugging

import SpecReduceUtils
import USERDefs


class vScrollFrame(tk.Frame):
  """
  Frame with a scroll bar for vertical
  """
  def __init__(self, parent, *args, **kwargs):
    tk.Frame.__init__(self,parent,*args,**kwargs)
    self.canvas =tk.Canvas(self, borderwidth=0)
    self.frame = tk.Frame(self.canvas)
    self.vscroll=tk.Scrollbar(self.canvas,orient="vertical",command=self.canvas.yview)
    self.canvas.configure(yscrollcommand=self.vscroll.set)
   
    self.vscroll.pack(side="right",fill="y")
    self.canvas.pack(side="left",fill="both",expand=True)
    self.canvas.create_window((0,0),window=self.frame,anchor="nw",tags="self.frame")
    self.frame.bind("<Configure>",self.onFrameConfigure)
    #self.frame.pack(fill="both")
  def onFrameConfigure(self,event):
    #print "DEBUG: running onFrameConfigure.   event:",event.height, "   ",event.width
    self.canvas.itemconfig(self.frame,height=event.height,width=event.width)
    #if (self.canvas.winfo_width() != self.frame.winfo_reqwidth() or self.canvas.winfo_height() !=self.frame.winfo_reqheight()):
      #print "DEBUG:running size branch  priorW:",self.canvas.winfo_width()," requestedW:",self.frame.winfo_reqwidth(),"    priorH:",self.canvas.winfo_height()," requestedH:",self.frame.winfo_reqheight()
    if (self.canvas.winfo_width() != self.canvas.winfo_reqwidth() or self.canvas.winfo_height() !=self.canvas.winfo_reqheight()):
    #if (self.canvas.winfo_width() > self.canvas.winfo_reqwidth() or self.canvas.winfo_height() > self.canvas.winfo_reqheight()):
      #print "DEBUG:running size branch  priorW:",self.canvas.winfo_width()," requestedW:",self.canvas.winfo_reqwidth(),"    priorH:",self.canvas.winfo_height()," requestedH:",self.canvas.winfo_reqheight()
      size = (self.frame.winfo_reqwidth(), self.frame.winfo_reqheight())
      self.canvas.config(scrollregion="0 0 %s %s" % size)
      #print "DEBUG:ran the size branch  priorW:",self.canvas.winfo_width()," requestedW:",self.canvas.winfo_reqwidth(),"    priorH:",self.canvas.winfo_height()," requestedH:",self.canvas.winfo_reqheight()

#pop up dialog class
class infoDialog(tkSimpleDialog.Dialog):
  def __init__(self,parent, info, filename ):
    #print "DEBUG: building infoDialog  args:"
    #for name, value in kwargs.items():
    #  print name,":  ",value
    self.filename=filename
    self.info=info
    #self.filename=kwargs[filename]
    #self.info=kwargs[info]
    self.keeping=True
    #tkSimpleDialog.Dialog.__init__(self,parent,*args,**kwargs)
    tkSimpleDialog.Dialog.__init__(self,parent)
  def body(self,master):  #change name to body for use
    self.TopLabel=tk.Label(master,text=self.filename)
    self.TopLabel.pack()
    self.tFrame=tk.Frame(master)
    self.tFrame.pack()
    self.scroll=tk.Scrollbar(self.tFrame)
    self.BigText=tk.Text(self.tFrame,height=50,width=100)
    self.scroll.pack(side=tk.RIGHT,fill=tk.Y)
    self.BigText.pack(side=tk.LEFT,fill=tk.Y)
    self.scroll.config(command=self.BigText.yview)
    self.BigText.config(yscrollcommand=self.scroll.set)
    self.BigText.insert(tk.END,self.info)
    self.BigText.config(state=tk.DISABLED)
  def lbody(self,master):
    self.TopLabel=tk.Label(master,text=self.filename)
    self.TopLabel.pack()
    self.BigText=tk.Label(master,text=self.info,bg='white',height=50,width=80)
    self.BigText.pack()
  def buttonbox(self):
    box=tk.Frame(self)
    removeButton = tk.Button(box,text="Remove",command=self.remove)
    removeButton.pack(side=tk.LEFT)
    keepButton = tk.Button(box,text="Keep",command=self.cancel,default=tk.ACTIVE)
    keepButton.pack(side=tk.LEFT)
    self.bind("<Return>",self.cancel)
    box.pack()
  def remove(self):
    self.keeping=False
    self.cancel()

#class specReduceGUI(object):
class specReduceGUI(tk.Frame):
  def __init__(self,parent,*args,**kwargs):
    tk.Frame.__init__(self,parent,*args,**kwargs)
    self.parent=parent
    self.parent.title("specReduceGUI")
    self.initialize()

  def initialize(self):
    """
    create buttons for reading and writting input files only.  Update function will
     handle setting up displayed values and displayed items will drive data
     interaction.  Also set up some data stuctures
    """
    self.Headers={}  #store header for each fits file already loaded.
    self.ReferenceSnips={} #store several lines of a reference file for display
    self.InputStruct={}
    self.InputStruct["names"]={}
    self.FrameByName={}
    self.FrameChildByName={}
    self.OptionVals={}
    self.RadioVals={}
    self.RadioValsCB={}
    self.AntiRadioStruct={}

    self.TopFrame=tk.Frame(self.parent,relief=tk.GROOVE,bd=5,width=1024)
    self.TopFrame.pack(side=tk.TOP,fill=tk.X)
    self.LoadInputButton=tk.Button(self.TopFrame,text="Load an input file",command=self.onLoadInput)
    self.LoadInputButton.pack(side=tk.LEFT)
    self.SaveInputButton=tk.Button(self.TopFrame,text="Save an input file",command=self.onSaveInput)
    self.SaveInputButton.pack(side=tk.LEFT)
    self.ExitButton=tk.Button(self.TopFrame,text="Exit specGUI",command=self.onExit)
    self.ExitButton.pack(side=tk.LEFT)

    self.AddItemsFrame=tk.Frame(self.parent,relief=tk.GROOVE,bd=5,width=1024)
    self.AddItemsFrame.pack(side=tk.TOP,fill=tk.X)
    self.InstructionLabel=tk.Label(self.AddItemsFrame,text="Click to create a new analysis element")
    self.InstructionLabel.pack(side=tk.TOP)
    self.AddNameLabel=tk.Label(self.AddItemsFrame,text="Name:")
    self.AddNameLabel.pack(side=tk.LEFT)
    self.NameVarGUI=tk.StringVar()
    self.NameEntryBox=tk.Entry(self.AddItemsFrame,textvariable=self.NameVarGUI)
    self.NameEntryBox.pack(side=tk.LEFT)
    self.AddBiasButton=tk.Button(self.AddItemsFrame,text="Bias",command=self.addBias)
    self.AddBiasButton.pack(side=tk.LEFT)
    self.AddFlatButton=tk.Button(self.AddItemsFrame,text="Flat",command=self.addFlat)
    self.AddFlatButton.pack(side=tk.LEFT)
    self.AddWLCalButton=tk.Button(self.AddItemsFrame,text="WL calibration",command=self.addWLCalibration)
    self.AddWLCalButton.pack(side=tk.LEFT)
    self.AddWLComButton=tk.Button(self.AddItemsFrame,text="WL combination",command=self.addWLCombination)
    self.AddWLComButton.pack(side=tk.LEFT)
    self.AddSpecButton=tk.Button(self.AddItemsFrame,text="Spectrum",command=self.addSpectrum)
    self.AddSpecButton.pack(side=tk.LEFT)
    #flux
    self.AddFluxButton=tk.Button(self.AddItemsFrame,text="Flux calibration",command=self.addFluxCal)
    self.AddFluxButton.pack(side=tk.LEFT)
    #telluric
    self.AddTelluricButton=tk.Button(self.AddItemsFrame,text="Telluric",command=self.addTelluric)
    self.AddTelluricButton.pack(side=tk.LEFT)
    #merge  
    self.AddMergeButton=tk.Button(self.AddItemsFrame,text="Merge Spectra",command=self.addMergeSpectrum)
    self.AddMergeButton.pack(side=tk.LEFT)

    #self.AnalysisRepresentFrame=tk.Frame(self.parent,relief=tk.GROOVE,bd=5,width=800)
    #self.AnalysisRepresentFrame=VerticalScrolledFrame(self.parent,relief=tk.GROOVE,bd=5,width=800)
    self.AnalysisRepresentFrame=vScrollFrame(self.parent,relief=tk.GROOVE,bd=5,width=800)
    #self.AnalysisRepresentFrame.pack(side=tk.TOP,fill=tk.X)
    self.AnalysisRepresentFrame.pack(side=tk.TOP,fill=tk.BOTH)


  def onExit(self):
    """
    exit specGUI
    """
    #TODO add checks to see if a save request should be prompted.
    if (tkMessageBox.askokcancel("Exit","Would you like to exit SpecGUI?")):
      self.parent.destroy()

  def onLoadInput(self):
    """
    get input file name, load it, and call update to set up environment.
    """
    print "DEBUG:loading"
    #loadTarget = tkFileDialog.askopenfilename(filetypes=[("input files","*")])
    loadTarget = tkFileDialog.askopenfilename()
    if (loadTarget):
      print "DEBUG: loading ",loadTarget
      tempInput=SpecReduceUtils.fromInput(loadTarget)
      self.InputStruct=tempInput
      self.updateGUI()

  def onSaveInput(self):
    """
    get name of save file, and write data to said file.  No need to update GUI.
    """
    print "DEBUG:saving"
    saveTarget = tkFileDialog.asksaveasfilename()
    if (saveTarget):
      print "DEBUG: saving ",saveTarget
      SpecReduceUtils.toInput(saveTarget,self.InputStruct)


  def updateGUI(self):
    """
    process data stored in the input data structure and create GUI elements to modify it
    """
    #define some internal helper functions
    def fillLabel(tkFrame,obj,startColumn=0,shiftdown=0):
      NameLabel=tk.Label(tkFrame,text=obj)
      NameLabel.grid(column=startColumn,row=1+shiftdown)
      return startColumn+1
    def fillDelete(tkFrame,obj,startColumn,shiftdown=0):
      DelButton=tk.Button(tkFrame,text="Delete",background="orange red",command=lambda obj=obj: self.deleteElement(obj))
      DelButton.grid(column=startColumn,row=1+shiftdown)
      return startColumn+1
    def fillBias(tkFrame,obj,startColumn,shiftdown=0):
      BiasLabel=tk.Label(tkFrame,text="Bias:")
      BiasLabel.grid(column=startColumn,row=1+shiftdown)
      if ("bias" not in self.OptionVals):
        self.OptionVals["bias"]={}
      self.OptionVals["bias"][obj]=tk.StringVar()
      if ("bias" in self.InputStruct):
        biases=[bias for bias in self.InputStruct["bias"]]
      else:
        biases=[]
      biases.insert(0,"*over")
      biases.insert(0,"*none")
      BiasOpts=tk.OptionMenu(tkFrame,self.OptionVals["bias"][obj],*biases)
      BiasOpts.grid(column=startColumn,row=2+shiftdown)
      self.OptionVals["bias"][obj].set(self.InputStruct[self.InputStruct["names"][obj]][obj]["bias"])
      self.OptionVals["bias"][obj].trace("w",lambda foo,bar,baz,obj=obj : self.onSelectChange("bias",obj,self.OptionVals["bias"][obj]))
      nextColumn=startColumn+1
      return nextColumn
    def fillFlat(tkFrame,obj,startColumn,shiftdown=0):
      FlatLabel=tk.Label(tkFrame,text="Flat")
      FlatLabel.grid(column=startColumn,row=1+shiftdown)
      if ("flat" not in self.OptionVals):
        self.OptionVals["flat"]={}
      self.OptionVals["flat"][obj]=tk.StringVar()
      if ("flat" in self.InputStruct):
        flats=[flat for flat in self.InputStruct["flat"]]
      else:
        flats=[]
      flats.insert(0,"*none")
      FlatOpts=tk.OptionMenu(tkFrame,self.OptionVals["flat"][obj],*flats)
      FlatOpts.grid(column=startColumn,row=2+shiftdown)
      self.OptionVals["flat"][obj].set(self.InputStruct[self.InputStruct["names"][obj]][obj]["flat"])
      self.OptionVals["flat"][obj].trace("w",lambda foo,bar,baz,obj=obj : self.onSelectChange("flat",obj,self.OptionVals["flat"][obj]))
      nextColumn=startColumn+1
      return nextColumn
    def fillFrames(tkFrame,obj,startColumn,numCols=4,shiftdown=0):
      #frames=SpecReduceUtils.fileListParser(self.InputStruct[self.InputStruct["names"][obj]][obj]["frame"])
      frames=self.InputStruct[self.InputStruct["names"][obj]][obj]["frame"]
      FrmLabel=tk.Label(tkFrame,text="Frames:")
      FrmLabel.grid(column=startColumn,row=1+shiftdown)
      #NewFrm=tk.Button(tkFrame,text="add fits",command=lambda obj=obj :self.addFile(self.InputStruct["names"][obj],obj,"frame","fits"))
      NewFrm=tk.Button(tkFrame,text="add fits",command=lambda obj=obj :self.addFits(self.InputStruct["names"][obj],obj,"frame","fits"))
      NewFrm.grid(column=startColumn,row=2+shiftdown)
      i,j=0,1+shiftdown
      imax=0
      #print "DEBUG: about to fill in ",len(frames)," frames"
      for frm in sorted(frames):
        #print "DEBUG:  filling frames for ",obj,"  current frame is ",frm
        imax=max(imax,i)
        drive,path=os.path.split(frm)
        path,filename=os.path.split(path)
        fButton=tk.Button(tkFrame,text=filename,command=lambda obj=obj,frm=frm: self.displayFileInfo(self.InputStruct["names"][obj],obj,"frame",frm))
        fButton.grid(column=startColumn+1+i,row=j)
        i+=1
        if (i>=numCols):
          i=0
          j+=1
      return startColumn+2+imax
    def fillRefs(tkFrame,obj,startColumn,numCols=4,shiftdown=0):
      #refs=SpecReduceUtils.fileListParser(self.InputStruct[self.InputStruct["names"][obj]][obj]["ref"])
      refs=self.InputStruct[self.InputStruct["names"][obj]][obj]["ref"]
      RefLabel=tk.Label(tkFrame,text="References:")
      RefLabel.grid(column=startColumn,row=1+shiftdown)
      NewRef=tk.Button(tkFrame,text="add ref",command=lambda obj=obj :self.addFile(self.InputStruct["names"][obj],obj,"ref","dat"))
      NewRef.grid(column=startColumn,row=1)
      i,j=0,1+shiftdown
      imax=0
      for ref in refs:
        imax=max(imax,i)
        drive,path=os.path.split(ref)
        path,filename=os.path.split(path)
        rButton=tk.Button(tkFrame,text=filename,command=lambda obj=obj,ref=ref: self.displayFileInfo(self.InputStruct["names"][obj],obj,"ref",ref,mode='ref'))
        rButton.grid(column=startColumn+1+i,row=j)
        i+=1
        if (i>=numCols):
          i=0
          j+=1
      return startColumn+2+imax
    def fillWLC(tkFrame,obj,startColumn,shiftdown=0):
      WLCLabel=tk.Label(tkFrame,text="WL Calibration:")
      WLCLabel.grid(column=startColumn,row=1+shiftdown)
      if ("WL_cal" not in self.OptionVals):
        self.OptionVals["WL_cal"]={}
      self.OptionVals["WL_cal"][obj]=tk.StringVar()
      if ("WL_calibrate" in self.InputStruct):
        WLCals=[WLC for WLC in self.InputStruct["WL_calibrate"]]
      else:
        WLCals=[]
      #need to add combos as well...
      if ("WL_combo" in self.InputStruct):
        WLCals.extend([WLC for WLC in self.InputStruct["WL_combo"]])
      WLCals.insert(0,"*none")
      WLCOpts=tk.OptionMenu(tkFrame,self.OptionVals["WL_cal"][obj],*WLCals)
      WLCOpts.grid(column=startColumn,row=2+shiftdown)
      self.OptionVals["WL_cal"][obj].set(self.InputStruct[self.InputStruct["names"][obj]][obj]["WL_cal"])
      self.OptionVals["WL_cal"][obj].trace("w",lambda foo,bar,baz,obj=obj : self.onSelectChange("WL_cal",obj,self.OptionVals["WL_cal"][obj]))
      nextColumn=startColumn+1
      return nextColumn
    def fillFluxCal(tkFrame,obj,startColumn,shiftdown=0):
      FCLabel=tk.Label(tkFrame,text="Flux Calibration:")
      FCLabel.grid(column=startColumn,row=1+shiftdown)
      if ("flux_calibrator" not in self.OptionVals):
        self.OptionVals["flux_calibrator"]={}
      self.OptionVals["flux_calibrator"][obj]=tk.StringVar()
      if ("flux_calibrate" in self.InputStruct):
        FCals=[FC for FC in self.InputStruct["flux_calibrate"]]
      else:
        FCals=[]
      FCals.insert(0,"*none")
      FCOpts=tk.OptionMenu(tkFrame,self.OptionVals["flux_calibrator"][obj],*FCals)
      FCOpts.grid(column=startColumn,row=2+shiftdown)
      self.OptionVals["flux_calibrator"][obj].set(self.InputStruct[self.InputStruct["names"][obj]][obj]["flux_calibrator"])
      self.OptionVals["flux_calibrator"][obj].trace("w",lambda foo,bar,baz,obj=obj : self.onSelectChange("flux_calibrator",obj,self.OptionVals["flux_calibrator"][obj]))
      nextColumn=startColumn+1
      return nextColumn
    def fillTelluric(tkFrame,obj,startColumn,shiftdown=0):
      TLabel=tk.Label(tkFrame,text="Telluric Correction:")
      TLabel.grid(column=startColumn,row=1+shiftdown)
      if ("telluric" not in self.OptionVals):
        self.OptionVals["telluric"]={}
      self.OptionVals["telluric"][obj]=tk.StringVar()
      if ("telluric_correct" in self.InputStruct):
        TCals=[TC for TC in self.InputStruct["telluric_correct"]]
      else:
        TCals=[]
      TCals.insert(0,"*none")
      TCOpts=tk.OptionMenu(tkFrame,self.OptionVals["telluric"][obj],*TCals)
      TCOpts.grid(column=startColumn,row=2+shiftdown)
      self.OptionVals["telluric"][obj].set(self.InputStruct[self.InputStruct["names"][obj]][obj]["telluric"])
      self.OptionVals["telluric"][obj].trace("w",lambda foo,bar,baz,obj=obj : self.onSelectChange("telluric",obj,self.OptionVals["telluric"][obj]))
      nextColumn=startColumn+1
      return nextColumn
    def fillSkyCal(tkFrame,obj,startColumn,shiftdown=0):
      SCLabel=tk.Label(tkFrame,text="Sky Calibration Correction:")
      SCLabel.grid(column=startColumn,row=1+shiftdown,columnspan=2)
      if ("skyCalCorrect" not in self.OptionVals):
        self.OptionVals["skyCalCorrect"]={}
      self.OptionVals["skyCalCorrect"][obj]=tk.StringVar()
      tk.Radiobutton(tkFrame,text="Yes",variable=self.OptionVals["skyCalCorrect"][obj],value="1").grid(column=startColumn,row=2+shiftdown)
      tk.Radiobutton(tkFrame,text="No",variable=self.OptionVals["skyCalCorrect"][obj],value="0").grid(column=startColumn+1,row=2+shiftdown)
      self.OptionVals["skyCalCorrect"][obj].set(self.InputStruct[self.InputStruct["names"][obj]][obj]["skyCalCorrect"])
      self.OptionVals["skyCalCorrect"][obj].trace("w",lambda foo,bar,baz,obj=obj : self.onSelectChange("skyCalCorrect",obj,self.OptionVals["skyCalCorrect"][obj]))
      return startColumn+2
    def fillSpec(tkFrame,obj,startColumn,shiftdown=0,label="Spectrum:",key="spectrum"):
      SpectrumLabel=tk.Label(tkFrame,text=label)
      SpectrumLabel.grid(column=startColumn,row=1+shiftdown)
      if (key not in self.OptionVals):
        self.OptionVals[key]={}
      self.OptionVals[key][obj]=tk.StringVar()
      if ("spectrum" in self.InputStruct):
        spectra=[spectrum for spectrum in self.InputStruct["spectrum"]]
      else:
        spectra=[]
      #pectra.insert(0,"*none")
      SpectrumOpts=tk.OptionMenu(tkFrame,self.OptionVals[key][obj],*spectra)
      SpectrumOpts.grid(column=startColumn,row=2+shiftdown)
      self.OptionVals[key][obj].set(self.InputStruct[self.InputStruct["names"][obj]][obj][key])
      self.OptionVals[key][obj].trace("w",lambda foo,bar,baz,obj=obj : self.onSelectChange(key,obj,self.OptionVals[key][obj]))
      nextColumn=startColumn+1
      return nextColumn
    def fillRef(tkFrame,obj,startColumn,shiftdown=0):
      ref = self.InputStruct[self.InputStruct["names"][obj]][obj]["reference"]
      ref=ref.strip("'")
      RefLabel=tk.Label(tkFrame,text="Reference:")
      RefLabel.grid(column=startColumn,row=1+shiftdown)
      NewRef=tk.Button(tkFrame,text="change ref",command=lambda obj=obj :self.changeFile(self.InputStruct["names"][obj],obj,"reference","dat"))
      NewRef.grid(column=startColumn,row=2+shiftdown)
      drive,path=os.path.split(ref)
      path,filename=os.path.split(path)
      rButton=tk.Button(tkFrame,text=filename,command=lambda obj=obj,ref=ref: self.displayFileInfo(self.InputStruct["names"][obj],obj,"reference",ref,mode='single'))
      rButton.grid(column=startColumn+1,row=1+shiftdown)
      nextColumn=startColumn+2
      return nextColumn
    def fillFiles(tkFrame,obj,startColumn,shiftdown=0,numCols=4):
      #files=SpecReduceUtils.fileListParser(self.InputStruct[self.InputStruct["names"][obj]][obj]["calibrationsets"])
      files=self.InputStruct[self.InputStruct["names"][obj]][obj]["calibrationsets"]
      FLabel=tk.Label(tkFrame,text="Calibrations:")
      FLabel.grid(column=startColumn,row=1+shiftdown)
      NewCal=tk.Button(tkFrame,text="add calibration",command=lambda obj=obj :self.addFile(self.InputStruct["names"][obj],obj,"reference","map"))
      NewCal.grid(column=startColumn,row=1)
      i,j=0,1+shiftdown
      imax=0
      #print "DEBUG: about to fill in ",len(frames)," frames"
      for fil in files:
        #print "DEBUG:  filling frames for ",obj,"  current frame is ",frm
        imax=max(imax,i)
        drive,path=os.path.split(fil)
        path,filename=os.path.split(path)
        fButton=tk.Button(tkFrame,text=filename,command=lambda obj=obj,fil=fil: self.displayFileInfo(self.InputStruct["names"][obj],obj,"reference",fil,mode='map'))
        fButton.grid(column=startColumn+1+i,row=j)
        i+=1
        if (i>=numCols):
          i=0
          j+=1
      return startColumn+2+imax
    def fillFile(tkFrame,obj,startColumn,shiftdown=0):
      File = self.InputStruct[self.InputStruct["names"][obj]][obj]["fromfile"]
      File = File.strip("'")
      FileLabel=tk.Label(tkFrame,text="Load File:")
      FileLabel.grid(column=startColumn,row=1+shiftdown)
      NewFile=tk.Button(tkFrame,text="change file",command=lambda obj=obj :self.changeFile(self.InputStruct["names"][obj],obj,"fromfile"))
      NewFile.grid(column=startColumn,row=2+shiftdown)
      drive,path=os.path.split(File)
      path,filename=os.path.split(path)
      fButton=tk.Button(tkFrame,text=filename,command=lambda obj=obj,File=File: self.displayFileInfo(self.InputStruct["names"][obj],obj,"fromfile",File,mode='single'))
      fButton.grid(column=startColumn+1,row=1+shiftdown)
      nextColumn=startColumn+2
      return nextColumn
    def fillRadio(tkFrame,obj,startColumn,initialValue,entryFunc,blankStruct,shiftdown=0):
      if (obj in self.RadioVals):
        self.RadioVals[obj].trace_vdelete("w",self.RadioValsCB[obj])
      else: 
        self.RadioVals[obj]=tk.BooleanVar()
      self.RadioVals[obj].set(initialValue)
      tk.Radiobutton(tkFrame,text="Create",variable=self.RadioVals[obj],value=False).grid(column=startColumn,row=1+shiftdown)
      tk.Radiobutton(tkFrame,text="Load",variable=self.RadioVals[obj],value=True).grid(column=startColumn,row=2+shiftdown)
      self.RadioValsCB[obj]=self.RadioVals[obj].trace("w",lambda foo,bar,bas,obj=obj : self.onRadioChange(obj,self.RadioVals[obj],entryFunc,blankStruct))
      return startColumn+1
        
    def addNote(frame,note,col=0,row=0):
      #this is reall similar to fillLabel.  I have separated it so that any future changes can be kept separate
      NoteLabel=tk.Label(frame,text=note)
      NoteLabel.grid(column=col,row=row,columnspan=22,sticky=tk.W)
      
      
    def entryBias(bias):
      if (bias in self.FrameChildByName):
        self.FrameChildByName[bias].destroy()
      self.FrameChildByName[bias] = tk.Frame(self.FrameByName[bias])
      self.FrameChildByName[bias].pack(side=tk.LEFT)
      nextColumn=fillLabel(self.FrameChildByName[bias],bias)
      nextColumn=fillDelete(self.FrameChildByName[bias],bias,nextColumn-1,shiftdown=1)
      if ("fromfile" in self.InputStruct["bias"][bias]):
        addNote(self.FrameChildByName[bias],"BIAS: Select a previously created bias file")
        nextColumn=fillRadio(self.FrameChildByName[bias],bias,nextColumn,True,entryBias,{"frame":"[]"})
        nextColumn=fillFile(self.FrameChildByName[bias],bias,nextColumn)
      else:
        addNote(self.FrameChildByName[bias],"BIAS: Specify bias frames to combine")
        nextColumn=fillRadio(self.FrameChildByName[bias],bias,nextColumn,False,entryBias,{"fromfile":""})
        nextColumn=fillFrames(self.FrameChildByName[bias],bias,nextColumn,5)
    def entryFlat(flat):
      if (flat in self.FrameChildByName):
        self.FrameChildByName[flat].destroy()
      self.FrameChildByName[flat] = tk.Frame(self.FrameByName[flat])
      self.FrameChildByName[flat].pack(side=tk.LEFT)
      nextColumn=fillLabel(self.FrameChildByName[flat],flat)
      nextColumn=fillDelete(self.FrameChildByName[flat],flat,nextColumn-1,shiftdown=1)
      if ("fromfile" in self.InputStruct["flat"][flat]):
        addNote(self.FrameChildByName[flat],"FLAT: Select a previously created flat field file")
        nextColumn=fillRadio(self.FrameChildByName[flat],flat,nextColumn,True,entryFlat,{"bias":"*none","frame":"[]"})
        nextColumn=fillRadio(self.FrameChildByName[flat],flat,nextColumn,False,entryFlat,{"fromfile":""})
        nextColumn=fillFile(self.FrameChildByName[flat],flat,nextColumn)
      else:
        addNote(self.FrameChildByName[flat],"FLAT: Specify bias method and frames to include in the flat field")
        nextColumn=fillRadio(self.FrameChildByName[flat],flat,nextColumn,False,entryFlat,{"fromfile":""})
        nextColumn=fillBias(self.FrameChildByName[flat],flat,nextColumn)
        nextColumn=fillFrames(self.FrameChildByName[flat],flat,nextColumn,6)
    def entryWLC(WLC):
      if (WLC in self.FrameChildByName):
        self.FrameChildByName[WLC].destroy()
      self.FrameChildByName[WLC] = tk.Frame(self.FrameByName[WLC])
      self.FrameChildByName[WLC].pack(side=tk.LEFT)
      nextColumn=fillLabel(self.FrameChildByName[WLC],WLC)
      nextColumn=fillDelete(self.FrameChildByName[WLC],WLC,nextColumn-1,shiftdown=1)
      if ("fromfile" in self.InputStruct["WL_calibrate"][WLC]):
        addNote(self.FrameChildByName[WLC],"WL CAL: Select a previously created wavelength calibration file")
        nextColumn=fillRadio(self.FrameChildByName[WLC],WLC,nextColumn,True,entryWLC,{"bias":"*none","flat":"*none","ref":"[]","frame":"[]"})
        nextColumn=fillFile(self.FrameChildByName[WLC],WLC,nextColumn)
      else:
        addNote(self.FrameChildByName[WLC],"WL CAL: Specify bias method (optional), flat field (optional), arc line list, and frames for wavelength calibration")
        nextColumn=fillRadio(self.FrameChildByName[WLC],WLC,nextColumn,False,entryWLC,{"fromfile":""})
        nextColumn=fillBias(self.FrameChildByName[WLC],WLC,nextColumn)
        nextColumn=fillFlat(self.FrameChildByName[WLC],WLC,nextColumn)
        nextColumn=fillRefs(self.FrameChildByName[WLC],WLC,nextColumn,2)
        nextColumn=fillFrames(self.FrameChildByName[WLC],WLC,nextColumn,3)
    def entryWLCombo(WLC):
      if (WLC in self.FrameChildByName):
        self.FrameChildByName[WLC].destroy()
      self.FrameChildByName[WLC] = tk.Frame(self.FrameByName[WLC])
      self.FrameChildByName[WLC].pack(side=tk.LEFT)
      addNote(self.FrameChildByName[WLC],"WL CAL (COMBINE): Select wavelength calibration files to combine and re-fit")
      nextColumn=fillLabel(self.FrameChildByName[WLC],WLC)
      nextColumn=fillDelete(self.FrameChildByName[WLC],WLC,nextColumn-1,shiftdown=1)
      nextColumn=fillFiles(self.FrameChildByName[WLC],WLC,nextColumn,4)
    def entrySpectrum(spec):
      if (spec in self.FrameChildByName):
        self.FrameChildByName[spec].destroy()
      self.FrameChildByName[spec] = tk.Frame(self.FrameByName[spec])
      self.FrameChildByName[spec].pack(side=tk.LEFT)
      nextColumn=fillLabel(self.FrameChildByName[spec],spec)
      nextColumn=fillDelete(self.FrameChildByName[spec],spec,nextColumn-1,shiftdown=1)
      if ("fromfile" in self.InputStruct["spectrum"][spec]):
        addNote(self.FrameChildByName[spec],"SPECTRUM: Select a previously reduced spectrum file")
        nextColumn=fillRadio(self.FrameChildByName[spec],spec,nextColumn,True,entrySpectrum,{"bias":"*none","flat":"*none","WL_cal":"*none","flux_calibrator":"*none","frame":"[]"})
        nextColumn=fillFile(self.FrameChildByName[spec],spec,nextColumn)
      else:
        addNote(self.FrameChildByName[spec],"SPECTRUM: Specify bias method, flat field, wavelength calibration (optional), sky line wavelength correction, flux calibration (optional), and frames to combine")
        nextColumn=fillRadio(self.FrameChildByName[spec],spec,nextColumn,False,entrySpectrum,{"fromfile":""})
        nextColumn=fillBias(self.FrameChildByName[spec],spec,nextColumn)
        nextColumn=fillFlat(self.FrameChildByName[spec],spec,nextColumn)
        nextColumn=fillWLC(self.FrameChildByName[spec],spec,nextColumn)
        nextColumn=fillSkyCal(self.FrameChildByName[spec],spec,nextColumn)
        nextColumn=fillFluxCal(self.FrameChildByName[spec],spec,nextColumn)
        nextColumn=fillTelluric(self.FrameChildByName[spec],spec,nextColumn)
        nextColumn=fillFrames(self.FrameChildByName[spec],spec,nextColumn,3)
    def entryFluxCal(flux):
      if (flux in self.FrameChildByName):
        self.FrameChildByName[flux].destroy()
      self.FrameChildByName[flux] = tk.Frame(self.FrameByName[flux])
      self.FrameChildByName[flux].pack(side=tk.LEFT)
      nextColumn=fillLabel(self.FrameChildByName[flux],flux)
      nextColumn=fillDelete(self.FrameChildByName[flux],flux,nextColumn-1,shiftdown=1)
      if ("fromfile" in self.InputStruct["flux_calibrate"][flux]):
        addNote(self.FrameChildByName[flux],"FLUX: Select a previously created flux calibration file")
        nextColumn=fillRadio(self.FrameChildByName[flux],flux,nextColumn,True,entryFluxCal,{"spectrum":"","reference":""})
        nextColumn=fillFile(self.FrameChildByName[flux],flux,nextColumn)
      else:
        addNote(self.FrameChildByName[flux],"FLUX: Specify a spectrum and a flux reference file")
        nextColumn=fillRadio(self.FrameChildByName[flux],flux,nextColumn,False,entryFluxCal,{"fromfile":""})
        nextColumn=fillSpec(self.FrameChildByName[flux],flux,nextColumn)
        nextColumn=fillRef(self.FrameChildByName[flux],flux,nextColumn)
    def entryTelluric(tell):
      if (tell in self.FrameChildByName):
        self.FrameChildByName[tell].destroy()
      self.FrameChildByName[tell] = tk.Frame(self.FrameByName[tell])
      self.FrameChildByName[tell].pack(side=tk.LEFT)
      nextColumn=fillLabel(self.FrameChildByName[tell],tell)
      nextColumn=fillDelete(self.FrameChildByName[tell],tell,nextColumn-1,shiftdown=1)
      if ("fromfile" in self.InputStruct["telluric_correct"][tell]):
        addNote(self.FrameChildByName[tell],"TELLURIC: Select a previously created telluric correction file")
        nextColumn=fillRadio(self.FrameChildByName[tell],tell,nextColumn,True,entryTelluric,{"spectrum":""})
        nextColumn=fillFile(self.FrameChildByName[tell],tell,nextColumn)
      else:
        addNote(self.FrameChildByName[tell],"TELLURIC: Specify a spectrum for telluric correction")
        nextColumn=fillRadio(self.FrameChildByName[tell],tell,nextColumn,False,entryTelluric,{"fromfile":""})
        nextColumn=fillSpec(self.FrameChildByName[tell],tell,nextColumn)
    def entryMergeSpectra(merg):
      if (merg in self.FrameChildByName):
        self.FrameChildByName[merg].destroy()
      self.FrameChildByName[merg] = tk.Frame(self.FrameByName[merg])
      self.FrameChildByName[merg].pack(side=tk.LEFT)
      addNote(self.FrameChildByName[merg],"MERGE: Select a blue and red spectrum to splice")
      nextColumn=fillLabel(self.FrameChildByName[merg],merg)
      nextColumn=fillDelete(self.FrameChildByName[merg],merg,nextColumn-1,shiftdown=1)
      nextColumn=fillSpec(self.FrameChildByName[merg],merg,nextColumn,label="Blue:",key="blue")
      nextColumn=fillSpec(self.FrameChildByName[merg],merg,nextColumn,label="Red:",key="red")

    self.AnalysisRepresentFrame.destroy()
    self.AnalysisRepresentFrame=vScrollFrame(self.parent,relief=tk.GROOVE,bd=5)
    self.AnalysisRepresentFrame.pack(side=tk.TOP,fill=tk.BOTH,expand=True)

    k=0
    #bias
    if ("bias" in self.InputStruct):
      for bias in self.InputStruct["bias"]:
        self.FrameByName[bias]=tk.Frame(self.AnalysisRepresentFrame.frame,bd=2,relief=tk.GROOVE)
        self.FrameByName[bias].grid(column=0,row=k,sticky=tk.W+tk.E)
        k+=1
        entryBias(bias)
    #flat
    if ("flat" in self.InputStruct):
      for flat in self.InputStruct["flat"]:
        self.FrameByName[flat]=tk.Frame(self.AnalysisRepresentFrame.frame,bd=2,relief=tk.GROOVE)
        self.FrameByName[flat].grid(column=0,row=k,sticky=tk.W+tk.E)
        k+=1
        entryFlat(flat)
    #WL_calibrate
    if ("WL_calibrate" in self.InputStruct):
      for WLC in self.InputStruct["WL_calibrate"]:
        self.FrameByName[WLC]=tk.Frame(self.AnalysisRepresentFrame.frame,bd=2,relief=tk.GROOVE)
        self.FrameByName[WLC].grid(column=0,row=k,sticky=tk.W+tk.E)
        k+=1
        entryWLC(WLC)
    #WL_combo
    if ("WL_combo" in self.InputStruct):
      for WLC in self.InputStruct["WL_combo"]:
        self.FrameByName[WLC]=tk.Frame(self.AnalysisRepresentFrame.frame,bd=2,relief=tk.GROOVE)
        self.FrameByName[WLC].grid(column=0,row=k,sticky=tk.W+tk.E)
        k+=1
        entryWLCombo(WLC)
    #spectrum
    if ("spectrum" in self.InputStruct):
      for spec in self.InputStruct["spectrum"]:
        self.FrameByName[spec]=tk.Frame(self.AnalysisRepresentFrame.frame,bd=2,relief=tk.GROOVE)
        self.FrameByName[spec].grid(column=0,row=k,sticky=tk.W+tk.E)
        k+=1
        entrySpectrum(spec)
    #flux_calibrate
    if ("flux_calibrate" in self.InputStruct):
      for flux  in self.InputStruct["flux_calibrate"]:
        self.FrameByName[flux]=tk.Frame(self.AnalysisRepresentFrame.frame,bd=2,relief=tk.GROOVE)
        self.FrameByName[flux].grid(column=0,row=k,sticky=tk.W+tk.E)
        k+=1
        entryFluxCal(flux)
    #telluric
    if ("telluric_correct" in self.InputStruct):
      for tell in self.InputStruct["telluric_correct"]:
        self.FrameByName[tell]=tk.Frame(self.AnalysisRepresentFrame.frame,bd=2,relief=tk.GROOVE)
        self.FrameByName[tell].grid(column=0,row=k,sticky=tk.W+tk.E)
        k+=1
        entryTelluric(tell)
    #mergeSpecs
    if ("mergeSpectra" in self.InputStruct):
      for merg in self.InputStruct["mergeSpectra"]:
        self.FrameByName[merg]=tk.Frame(self.AnalysisRepresentFrame.frame,bd=2,relief=tk.GROOVE)
        self.FrameByName[merg].grid(column=0,row=k,sticky=tk.W+tk.E)
        k+=1
        entryMergeSpectra(merg)

  #top level interaction elements
  def inputFiller(self,objtype,obj,elem,val):
    if (objtype not in self.InputStruct):
      self.InputStruct[objtype]={}
    if (obj not in self.InputStruct[objtype]):
      self.InputStruct[objtype][obj]={}
    if (elem not in self.InputStruct[objtype][obj]):
      self.InputStruct[objtype][obj][elem]=None
    self.InputStruct[objtype][obj][elem]=val
    
  def addBias(self):
    """
    create a new bias analysis
    """  
    name=self.NameVarGUI.get()
    #print "DEBUG: stub   Add an object named ", name
    self.InputStruct["names"][name]="bias"
    self.inputFiller("bias",name,"frame",[])
    self.updateGUI()

  def addFlat(self):
    """
    create a new flat analysis
    """  
    name=self.NameVarGUI.get()
    #print "DEBUG: stub   Add an object named ", name
    self.InputStruct["names"][name]="flat"
    self.inputFiller("flat",name,"bias","*none")
    self.inputFiller("flat",name,"frame",[])
    self.updateGUI()

  def addWLCalibration(self):
    """
    create a new Wavelength calibration
    """  
    name=self.NameVarGUI.get()
    self.InputStruct["names"][name]="WL_calibrate"
    self.inputFiller("WL_calibrate",name,"bias","*none")
    self.inputFiller("WL_calibrate",name,"flat","*none")
    self.inputFiller("WL_calibrate",name,"ref",[])
    self.inputFiller("WL_calibrate",name,"frame",[])
    self.updateGUI()

  def addWLCombination(self):
    """
    allow user to create a new WL calibration from multiple WL calibrations (either through median combine or combining pixel-wavelength pairs)
    """
    name=self.NameVarGUI.get()
    #TODO  - actually impliment combo
    self.InputStruct["names"][name]="WL_combo"
    self.inputFiller("WL_combo",name,"WL_cal",[])
    self.updateGUI()

  def addSpectrum(self):
    """
    create a new spectrum extraction
    """
    name=self.NameVarGUI.get()
    self.InputStruct["names"][name]="spectrum"
    self.inputFiller("spectrum",name,"bias","*none")
    self.inputFiller("spectrum",name,"flat","*none")
    self.inputFiller("spectrum",name,"WL_cal","*none")
    self.inputFiller("spectrum",name,"flux_calibrator","*none")
    self.inputFiller("spectrum",name,"frame",[])
    self.updateGUI()

  def addFluxCal(self):
    """
    create a new flux calibration from a spectrum
    """
    name=self.NameVarGUI.get()
    self.InputStruct["names"][name]="flux_calibrate"
    self.inputFiller("flux_calibrate",name,"spectrum","")
    self.inputFiller("flux_calibrate",name,"reference","")
    self.updateGUI()

  def addTelluric(self):
    """
    create a new telluric correction from a spectrum
    """
    name=self.NameVarGUI.get()
    self.InputStruct["names"][name]="telluric_correct"
    self.inputFiller("telluric_correct",name,"spectrum","")
    self.updateGUI()

  def addMergeSpectrum(self):
    """
    create a new full range spectrum from red and blue spectra
    """
    name=self.NameVarGUI.get()
    self.InputStruct["names"][name]="mergeSpectra"
    self.inputFiller("mergeSpectra",name,"blue","")
    self.inputFiller("mergeSpectra",name,"red","")
    self.updateGUI()

  def deleteElement(self,name):
    """
    remove an analysis element
    """
    elemType=self.InputStruct["names"][name]
    if tkMessageBox.askyesno("Delete?","Do you want to delete "+elemType+" "+ name+" from the specReduce input?"):
      self.FrameByName[name].destroy()
      del self.InputStruct[elemType][name]
      del self.InputStruct["names"][name]
      self.updateGUI()

  #second level interaction elements (call backs will be lambdas specifying keys
  def onRadioChange(self,name,var,entryFunc,blankStruct):
    elemType= self.InputStruct["names"][name]
    buildingVal= "fromfile" in self.InputStruct[elemType][name]
    print "DEBUG radio triggered, radio button val:",var.get(),"   loadstate:",buildingVal
    if (buildingVal != var.get()):
      #the value of the fromfile radiobutton has flipped.
      print "DEBUG:  changing radion from ",buildingVal," to ",var.get(),"  for entry ",name,".   replacement blank is ",blankStruct
      tempstruct=self.InputStruct[self.InputStruct["names"][name]][name]
      self.InputStruct[self.InputStruct["names"][name]][name] ={}
      if (name in self.AntiRadioStruct):
        print "DEBUG:  using antiRadio structure"
        self.InputStruct[self.InputStruct["names"][name]][name]=self.AntiRadioStruct[name]
      else:
        self.InputStruct[elemType][name]=blankStruct
      self.AntiRadioStruct[name]=tempstruct
      entryFunc(name)

  def onSelectChange(self,elem,name,var):
    """
    if the selected Element has been updated, change data structure and don't redraw GUI
    """
    elemType=self.InputStruct["names"][name]
    #print "DEBUG: option of interest is ",elem,",",name,",",var.get()
    #print "DEBUG:  attempting to modify an option  before it was:",self.InputStruct[elemType][name][elem], "   now it is:",var.get()
    if (self.InputStruct[elemType][name][elem] != var.get()):
      #print "DEBUG: ",elemType," ",name," ",elem," is changing from ",self.InputStruct[elemType][name][elem]," to ",var.get()
      self.InputStruct[elemType][name][elem] = var.get()

  def addFits(self,elemtype,name,subelem,filterStr=None):
    """
    wrapper for addFile that checks the 'fpa' field of the header and adds it to the element, if needed
    """
    self.addFile(elemtype,name,subelem,filterStr)
    targ = self.InputStruct[elemtype][name][subelem][-1]
    rawFits = fits.open(targ)
    head = rawFits[0].header
    if ('FPA' in head):
      fpa=head['FPA']
      fpa=fpa.strip().lower()
      self.InputStruct[elemtype][name]['fpa']=fpa
    else:
      warnings.warn("WARNING FPA card for file "+str(f)+" is absent.",UserWarning)

  def addFile(self,elemtype,name,subelem,filterStr=None):
    """
    add a file (normally ref or fits) to an analysis object
    """
    #callbacks will be a lambda to specify entry type and name.
    #addTarget = tkFileDialog.askopenfilename()
    addTarget = tkFileDialog.askopenfilenames()
    if (addTarget):
      print "DEBUG: adding ",addTarget
      #elemString=self.InputStruct[elemtype][name][subelem]
      #listForm=SpecReduceUtils.fileListParser(elemString)
      #listForm.append(addTarget)
      #newString=SpecReduceUtils.fileListBuilder(listForm)
      #self.InputStruct[elemtype][name][subelem]=newString
      for targ in self.parent.tk.splitlist(addTarget):
        self.InputStruct[elemtype][name][subelem].append(targ)
      #self.InputStruct[elemtype][name][subelem].append(addTarget)
      self.updateGUI()

  def changeFile(self,elemtype,name,subelem,filterStr=None):
    """
    change a file (normally ref or fits) to an analysis object
    """
    #callbacks will be a lambda to specify entry type and name.
    chTarget = tkFileDialog.askopenfilename()
    if (chTarget):
      print "DEBUG: changing to ",chTarget
      self.InputStruct[elemtype][name][subelem]=chTarget
      self.updateGUI()

#  def addFits(self,elemtype,name,subelem):
#    """
#    add a fits file to an analysis element
#    """
#    print "DEBUG: stub"
#    #callbacks will be a lambda to specify entry type and name.
#
#  def addRef(self,elemtype,name,subelem):
#    """
#    add a reference file to an analysis element
#    """
#    print "DEBUG: stub"
#    #callbacks will be a lambda to specify entry type and name.
#
  def displayFileInfo(self,elemtype,name,subelem,filename,mode='fits'):
    """
    create an information and interaction popup for a file entry
    """
    print "DEBUG: create a popup that interacts with the file ",filename," of type ",subelem," from element ",name," which is a ",elemtype
    #callbacks will be a lambda to specify entry type and name.
    #  will create option to delete a file from the list used to obtain the info...

    #get info text
    if (filename in self.ReferenceSnips):
      #see if the infotext was cached from a previous viewing
      infotext=self.ReferenceSnips[filename]
    else:
      if (mode == 'fits'):
        #open file, get header
        rawFits = fits.open(filename)
        infotext=repr(rawFits[0].header)
        self.ReferenceSnips[filename]=infotext
      else:
        #not a fits, just display a few lines...
        counter=0
        mx=30
        infotext=""
        with open(filename,'r') as f:
          for line in f:
            infotext = infotext+line
            counter+=1
            if (counter >= mx):
              break
    popup = infoDialog(self,filename=filename,info=infotext)
    if (not popup.keeping):
      if (mode=='single'):
        print "please use change file button instead"
        return
      #elemString=self.InputStruct[elemtype][name][subelem]
      #listForm=SpecReduceUtils.fileListParser(elemString)
      #listForm.remove(filename)
      self.InputStruct[elemtype][name][subelem].remove(filename)
      #newString=SpecReduceUtils.fileListBuilder(listForm)
      #self.InputStruct[elemtype][name][subelem]=newString
      self.updateGUI()




if __name__ == "__main__":
  parent=tk.Tk()
  parent.geometry("1600x1200")
  app = specReduceGUI(parent)
  parent.mainloop()
