#!/usr/bin/env python


import Tkinter as tk
#import tkFileDialog
import tkMessageBox
#import tkSimpleDialog
import sys
import getopt
import os
from astropy.io import fits
import warnings
import time #debugging
import numpy as np
import json
import copy
import matplotlib.pyplot as plt

import SpecReduceUtils
import USERDefs

decosmic_ctrl=True#make false to speed up testing by skipping cosmic ray removal

def main(argv):
  inputfile = ''
  outputdir = ''
  calibrationmode='auto'
  extractmode    ='auto'
  try:
    opts, args = getopt.getopt(argv,"hi:o:m:e:",["indir=","output=","mode=","calibrationmode=","extract=","extractmode=","quick","debug","help"])
  except getopt.GetoptError, exc:
    print 'specReduce.py -i <infile> -o <outdir> -m <calibrationmode> -e <extractmode> --quick --debug'
    print "DEBUG: ",exc.msg
    sys.exit(2)
  for opt, arg in opts:
    if opt in ('-h',"--help"):
      print 'specReduce.py -i <infile> -o <outdir> -m <calibrationmode> -e <extractmode> --quick --debug'
      sys.exit()
    elif opt in ("-i", "--indir"):
      inputfile = arg
    elif opt in ("-o", "--output"):
      outputdir = arg
    elif opt in ("--quick", "-q"):
      print "Running quick reduction.  Cosmic rays will not be removed"
      global decosmic_ctrl
      decosmic_ctrl=False
    elif opt in ("--debug", "-d"):
      print "Running in debug mode.  additional outputs will be created"
      USERDefs.debug=True
    elif opt in ("-m", "--mode", "--calibrationmode"):
      if arg not in ('auto','manual'):
        print 'acceptible calibrationmode values are auto and manual'
        return
      calibrationmode= arg
    elif opt in ("-e", "--extract", "--extractmode"):
      if arg not in ('auto','manual'):
        print 'acceptible calibrationmode values are auto and manual'
        return
      extractmode= arg
    else:
      assert False, "unhandled option on command line"
  print 'Input file is "', inputfile,'"'
  print 'Output directory is "', outputdir,'"'
  inputStruct = SpecReduceUtils.fromInput(inputfile)

  productsStruct={}
  if (USERDefs.debug):
    print "value of dcosmic: ",decosmic_ctrl

  #make outputdir
  if not os.path.exists(outputdir):
    os.makedirs(outputdir)
  #set up frame defaults
  FrmSetup={'dbsp_blue':USERDefs.DBSPBlue,'dbsp_red2':USERDefs.DBSPRed}
  populateBiases(inputStruct,productsStruct,outputdir,FrmSetup)
  populateFlats(inputStruct,productsStruct,outputdir,FrmSetup)
  populateWLCals(inputStruct,productsStruct,outputdir,FrmSetup,calibrationmode)
  populateWLCombo(inputStruct,productsStruct,outputdir) #,FrmSetup)
  populateTelluric(inputStruct,productsStruct,outputdir,FrmSetup)
  populateFluxCals(inputStruct,productsStruct,outputdir,FrmSetup)
  populateSpectra(inputStruct,productsStruct,outputdir,FrmSetup,extractmode)
  #applyWLCals(inputStruct,productsStruct,outputdir,FrmSetup)   #originally thought I would need this, but i think it's rolled into spectra.  leaving this here for now as a reminder
  #applyTellurics(inputStruct,productsStruct,outputdir,FrmSetup)  #originally thought I would need this, but i think it's rolled into spectra.  leaving this here for now as a reminder
  #applyFluxCals(inputStruct,productsStruct,outputdir,FrmSetup)  #originally thought I would need this, but i think it's rolled into spectra.  leaving this here for now as a reminder
  mergeSpectra(inputStruct,productsStruct,outputdir,FrmSetup)
  return 1

def setupElementDir(outputdir,elem):
  elemdir = os.path.join(outputdir,elem)
  if os.path.exists(elemdir):
    warnings.warn("WARNING output directory "+str(elemdir)+" already exists.  This could cause overwrites and/or confusion",UserWarning)
  else:
    os.makedirs(elemdir)
  return elemdir

def getBias(inputStruct,productsStruct,elemtype,elem):
  bias=None
  if ('bias' not in inputStruct[elemtype][elem]):
    return bias
  if (inputStruct[elemtype][elem]['bias'] == "*over"):
    bias="*over"
  elif (inputStruct[elemtype][elem]['bias'] != "*none"):
    bias=productsStruct[inputStruct[elemtype][elem]['bias']]['fits']
  return bias

def getFlat(inputStruct,productsStruct,elemtype,elem):
  flat=None
  if ('flat' not in inputStruct[elemtype][elem]):
    return flat
  if (inputStruct[elemtype][elem]['flat'] != "*none"):
    flat=productsStruct[inputStruct[elemtype][elem]['flat']]['fits']
  return flat

def getCal(inputStruct,productsStruct,elemtype,elem):
  cal=None
  if ('WL_cal' not in inputStruct[elemtype][elem]):
    return cal
  if (USERDefs.debug):
    print "DEBUG  in getCal  WL_cal: ",inputStruct[elemtype][elem]['WL_cal']
  if (inputStruct[elemtype][elem]['WL_cal'] != "*none"):
    cal = productsStruct[inputStruct[elemtype][elem]['WL_cal']]['calibration']
  return cal

def getTell(inputStruct,productsStruct,elemtype,elem,WLcal=None):
  tell=None
  if ('telluric' not in inputStruct[elemtype][elem]):
    return tell
  if (inputStruct[elemtype][elem]['telluric'] != "*none"):
    tell = productsStruct[inputStruct[elemtype][elem]['telluric']]['telluric_correct']
    # if the input element has a WL cal assocated with it, interpolate the correction onto those wavelengths
    if (WLcal is not None):
      tell = np.interp(WLcal,productsStruct[inputStruct[elemtype][elem]['telluric']]['calibration'],tell)
  return tell

def getFCal(inputStruct,productsStruct,elemtype,elem,WLcal=None):
  fCal=None
  if ('flux_calibrator' not in inputStruct[elemtype][elem]):
    return fCal
  if (inputStruct[elemtype][elem]["flux_calibrator"] != "*none"):
    fCal = productsStruct[inputStruct[elemtype][elem]["flux_calibrator"]]['fluxCal']
    # if the input element has a WL cal assocated with it, interpolate the fCal onto those wavelengths
    if (WLcal is not None):
      fCal = np.interp(WLcal,productsStruct[inputStruct[elemtype][elem]['flux_calibrator']]['WLcal'],fCal)
  return fCal
  
  

def populateBiases(inputStruct,productsStruct,outputdir,FrmSetup):
  if ('bias' not in inputStruct):
    return
  for bias in inputStruct['bias']:
    biasdir = setupElementDir(outputdir,bias)
    productsStruct[bias] = processBias(bias,inputStruct['bias'][bias],biasdir,FrmSetup)

def processBias(name,biasElem,biasdir,FrmSetup):
  biasStruct={}
  biases=[]
  if "fromfile" in biasElem:
    biasStruct['fits']=SpecReduceUtils.Frame(biasElem['fromfile'])
    return biasStruct
  for f in biasElem['frame']:
    print "loading ",f," as a bias frame."
    biases.append(SpecReduceUtils.Frame(f,type='bias',savePath=biasdir,frameSetup=FrmSetup[biasElem['fpa']]))
  bias = SpecReduceUtils.frameMedian(biases)
  bias.toFile(os.path.join(biasdir,name+".fits"))
  biasStruct['fits']=bias
  return biasStruct

def populateFlats(inputStruct,productsStruct,outputdir,FrmSetup):
  if ('flat' not in inputStruct):
    return
  for flat in inputStruct['flat']:
    flatdir = setupElementDir(outputdir,flat)
    bias=getBias(inputStruct,productsStruct,'flat',flat)
    productsStruct[flat] = processFlat(flat,inputStruct['flat'][flat],flatdir,FrmSetup[inputStruct['flat'][flat]['fpa']],bias)

def processFlat(name,flatElem,flatdir,FrmSetup,bias):
  flatStruct={}
  flats=[]
  if "fromfile" in flatElem:
    flatStruct['fits']=SpecReduceUtils.Frame(flatElem['fromfile'])
    return flatStruct
  for f in flatElem['frame']:
    print "loading ",f," as a flat frame."
    flats.append(SpecReduceUtils.Frame(f,type='flat',savePath=flatdir,frameSetup=FrmSetup,bias=bias))
  median = SpecReduceUtils.frameMedian(flats)
  median.toFile(os.path.join(flatdir,name+"_median.fits"))
  respFrame=SpecReduceUtils.response(median,countThreshold=1000)
  respFrame.toFile(os.path.join(flatdir,name+"_response.fits"))
  flat=SpecReduceUtils.frameDivision(median,respFrame)
  flat.type='flat'
  flat.toFile(os.path.join(flatdir,name+".fits"))
  flatStruct['fits']=flat
  return flatStruct
    
def populateWLCals(inputStruct,productsStruct,outputdir,FrmSetup,mode='auto'):
  if ('WL_calibrate' not in inputStruct):
    return
  for wlc in inputStruct['WL_calibrate']:
    caldir = setupElementDir(outputdir,wlc)
    bias=getBias(inputStruct,productsStruct,'WL_calibrate',wlc)
    flat=getFlat(inputStruct,productsStruct,'WL_calibrate',wlc)
    frmsetup=None
    if ('fpa' in inputStruct['WL_calibrate'][wlc]):
      frmsetup=FrmSetup[inputStruct['WL_calibrate'][wlc]['fpa']]
    productsStruct[wlc]=processWLC(wlc,inputStruct['WL_calibrate'][wlc],caldir,frmsetup,bias,flat,mode)

def processWLC(name,calElem,caldir,frmsetup=None,bias=None,flat=None,mode='auto'):
  calStruct={} #this one will have array(calibration) and tuples(peakmap), array only is possible if loaded from a file
  specs=[]
  if "fromfile" in calElem:
    File = ''
    with open(calElem['fromfile'],'r') as f:
      File = f.read()
    if (File[0]=='{'):
      #json file, pairs specified
      peakmap = json.loads(File)['peakmap']
      if (USERDefs.debug):
        print "DEBUG: peakmap loaded: ",peakmap.keys()
      calStruct['peakmap']={}
      calStruct['peakmap']['peaks']=peakmap['peaks']
      calStruct['peakmap']['size']=peakmap['size']
      if (USERDefs.debug):
        print "DEBUGGING cal from file  input pairs: ",peakmap['peaks']
      calibration = SpecReduceUtils.pairs2Calibration(peakmap['peaks'],peakmap['size'])
      calStruct['calibration']=np.array(calibration)
      return calStruct
    else:
      #not a json list, assume it's a full calibration
      calibration = SpecReduceUtils.loadlist(f)
      calStruct['calibration']=np.array(calibration)
      return calStruct
  thetaConstant=True
  grateConstant=True
  theta=None
  grate=None
  for f in calElem['frame']:
    print "loading ",f," as a wavelength calibrator."
    frm=SpecReduceUtils.Frame(f,type='lamp',savePath=caldir,frameSetup=frmsetup,bias=bias)
    if (theta == None):
      theta = SpecReduceUtils.radiansFromString(frm.Head['angle'])
    if (grate == None):
      grate = SpecReduceUtils.gratingFromString(frm.Head['grating'])
    thetaConstant = thetaConstant and (theta == SpecReduceUtils.radiansFromString(frm.Head['angle']))
    grateConstant = grateConstant and (grate == SpecReduceUtils.gratingFromString(frm.Head['grating']))
    if flat is not None:
      frame=SpecReduceUtils.frameDivision(frm,flat)
      frame.toFile(os.path.join(caldir,f+"_flattened.fits"))
    else:
      frame=frm
    spec = SpecReduceUtils.simpleExtract(frm.Data)
    SpecReduceUtils.savelist(spec,os.path.join(caldir,f+"_simpleSpectrum.dat"))
    specs.append(spec)
  if (not (thetaConstant and grateConstant)):
    raise Exception("internally inconsistant calibration files.  please check that all wavelength calibration images in the set have same angle and grating.")
  calspectrum=np.median(specs,0)
  SpecReduceUtils.savelist(calspectrum,os.path.join(caldir,name+"_simpleSpectrum.dat"))
  foundPeaks = SpecReduceUtils.findPeaks(calspectrum,spectralFitSpan=frmsetup['illuminatedSpan'])
  SpecReduceUtils.savelist(foundPeaks,os.path.join(caldir,name+"_extractedLampPeaks.dat"))
  print "Loading reference files ",calElem['ref']
  referencePeaks = SpecReduceUtils.loadLamps(calElem['ref'])
  initialPairs=[]
  if (mode=='manual' or mode=='interactive'):
    print "use user selected peak,wavelength pairs to calibrate"
    #if (True):
    if (False):
      if (calElem['fpa']=='dbsp_blue'):
        #initialPairs = {2436: 5269.5366, 878: 3576.6156, 1007: 3718.2065, 1137: 3859.9114, 2076: 4879.8635} #broad initial set blue
        initialPairs = {1809.137698584545: 4589.8978, 2531.0309539176155: 5371.4892, 878.31830229442016: 3576.6156, 2436.6818434648021: 5269.5366, 1136.7754646496587: 3859.9114, 1827.2333502543042: 4609.5673, 1799.4343151799262: 4579.3495, 1007.5819349914012: 3718.2065, 2490.8854874975973: 5328.0376, 2075.9831542716033: 4879.8635, 746.87316863215881: 3443.8762} #initial Kirby Blue
      elif (calElem['fpa']=='dbsp_red2'):
        #initialPairs = {2632: 7635.105, 2744: 7723.8, 1123: 6402.246, 1814: 6965.43, 1462: 6678.2}  #red broad set
        initialPairs = {1814.3703009527144: 6965.43, 1770.2174413479941: 6929.468, 2632.2565410177367: 7635.105, 1122.5766974065721: 6402.246, 2741.0490900643949: 7723.8, 1896.3623493243344: 7032.4127, 1461.9064161875235: 6678.2, 1936.9765315803897: 7065.188} #kirby red initial
      else:
        print "DEBUG:  in John's testing shortcut, your FPA makes no sense"
    else:
      print "Mark peaks from reference files ",calElem['ref']
      initialPairs = SpecReduceUtils.interactiveCalibrationCollector(calspectrum,framepitch=frm.pitch,refPeaks=referencePeaks)
  elif (mode=='auto'):
    warnings.warn("WARNING automatic wavelength calibration has proven difficult.  Results might not be reliable at this point.",UserWarning)
    dispersionGuess = SpecReduceUtils.computeDispersion(frmsetup['fixedAngle'],theta,frmsetup['pixelSize'],frmsetup['focalLength'],grate)
    centralLambda = SpecReduceUtils.computeCentralWL(theta,grate,frmsetup['fixedAngle'])
    if (USERDefs.debug):
      print "variables for initial calibration:  angle:",frmsetup['fixedAngle']," pixelSize:",frmsetup['pixelSize']," focalLength:",frmsetup['focalLength']," grating:",grate," theta:",theta
    print "approx dispersion as computed from properties: ",dispersionGuess, "   and central WL: ",centralLambda
    initialCalibration = SpecReduceUtils.initialCalibration(dispersionGuess,centralLambda,len(calspectrum))
    SpecReduceUtils.savelist(initialCalibration,os.path.join(caldir,name+"_initialCalibration.dat"))
    #debug
    sortP = sorted(foundPeaks)
    initCalPeaks = initialCalibration[sortP]
    if (USERDefs.debug):
      print "Rough calibration of found peaks from lamp image:\n",initCalPeaks
    if (USERDefs.debug):
      print "reference peaks:\n",referencePeaks
    #end debug
    initialPairs = SpecReduceUtils.calibrateWavelength(foundPeaks,referencePeaks,initialCalibration)
    if (USERDefs.debug):
      print "result of correlation calibration:",initialPairs
  else:
    warnings.warn("WARNING unrecognized calibration mode "+mode+" in calibration element "+name+".  This Element will not be used.  Could cause later crash.",UserWarning)
  if (USERDefs.debug):
    print "DEBUG: initial calibration set: ",initialPairs
  finalPairs = SpecReduceUtils.expandCalibration(initialPairs,foundPeaks,referencePeaks)
  peakmap={'size':len(calspectrum),'peaks':finalPairs}
  calStruct['peakmap']=peakmap
  calibration = SpecReduceUtils.pairs2Calibration(finalPairs,len(calspectrum))
  SpecReduceUtils.savelist(calibration,os.path.join(caldir,name+"_calibration.dat"))
  calStruct['calibration']=calibration
  with open(os.path.join(caldir,name+"_calibration.json"), 'w') as f:
    outStruct=copy.deepcopy(calStruct)
    outStruct['calibration']=calStruct['calibration'].tolist()
    json.dump(outStruct,f,indent=2)
  return calStruct
    
  
def populateWLCombo(inputStruct,productsStruct,outputdir):  #,FrmSetup):
  if ('WL_combo' not in inputStruct):
    return
  for combo in inputStruct['WL_combo']:
    combodir = setupElementDir(outputdir,combo)
    productsStruct[combo]=processWLCombo(combo,productsStruct,inputStruct['WL_combo'][combo],combodir)

def processWLCombo(name,productsStruct,comboElem,combodir):
  #no from file option
  comboStruct={}
  lenconstant=True
  length=None
  allPairs={} 
  for peakset in comboElem['calibrationsets']:
    peakmap = productsStruct[peakset]['peakmap']
    print "DEBUG:  peakset: ",peakset," productsStruct: ",productsStruct[peakset],"   peakmap: ",peakmap
    if (length==None):
      if (USERDefs.debug):
        print "DEBUG: having problems with peakmap: ",peakmap
      length=peakmap['size']
    lenconstant= lenconstant and length==peakmap['size']
    allPairs.update(peakmap['peaks'])
  if (not lenconstant):
    warnings.warn("WARNING:  All wavelegth calibration sets for a wavelength combo must have the same length.  Returning without creating WLCombo "+name+".",UserWarning)
    return
  peakmap={'size':length,'peaks':allPairs}
  comboStruct['peakmap']=peakmap
  if (USERDefs.debug):
    print "DEBUG:  combining cals, peakmap: ",allPairs
  calibration = SpecReduceUtils.pairs2Calibration(allPairs,length)
  SpecReduceUtils.savelist(calibration,os.path.join(combodir,name+"_calibration.dat"))
  comboStruct['calibration']=calibration
  with open(os.path.join(combodir,name+"_calibration.json"), 'w') as f:
    outStruct=copy.deepcopy(comboStruct)
    outStruct['calibration']=comboStruct['calibration'].tolist()
    json.dump(outStruct,f,indent=2)
  return comboStruct

def populateSpectra(inputStruct,productsStruct,outputdir,FrmSetup,mode):
  ##DEBUG
  #for s in inputStruct['spectrum']:
  #  print "DEBUG: ",s," is a spectrum"
  ##/DEBUG
  if (mode=="auto"):
    interactive=False
  else:
    interactive=True
  if ('spectrum' not in inputStruct):
    print "No spectra found in control structure.  Not populating spectra."
    return
  for spec in inputStruct['spectrum']:
    print "DEBUG: populating spectrum: ",spec
    if (spec in productsStruct):
      print "spectrum "+spec+" already populated.  This is normal if it's used for Telluric correction or a Flux calibrator."
      continue
    specdir = setupElementDir(outputdir,spec)
    bias=getBias(inputStruct,productsStruct,'spectrum',spec)
    flat=getFlat(inputStruct,productsStruct,'spectrum',spec)
    cal =getCal( inputStruct,productsStruct,'spectrum',spec)
    tell=getTell(inputStruct,productsStruct,'spectrum',spec,cal)
    flux=getFCal(inputStruct,productsStruct,'spectrum',spec,cal)
    frmsetup=None
    if ('fpa' in inputStruct['spectrum'][spec]):
      frmsetup=FrmSetup[inputStruct['spectrum'][spec]['fpa']]
    productsStruct[spec]=processSpectrum(spec,inputStruct['spectrum'][spec],specdir,frmsetup,bias,flat,cal,interactive,tell,flux)

def processSpectrum(name,specElem,specdir,framesetup=None,bias=None,flat=None,cal=None,mode=None,telluric=None,fluxCal=None):
  #print "DEBUG: spectrum processing inputs: cal:",cal
  print "Working on spectrum: ",name
  specStruct={}
  if "fromfile" in specElem:
    File = ''
    with open(specElem['fromfile'],'r') as f:
      File = f.read()
      #json file, pairs specified
    jStruct = json.loads(File)
    specStruct['spectrum']=np.array(jStruct['spectrum'])
    specStruct['skyspectrum']=np.array(jStruct['skyspectrum'])
    specStruct['variancespectrum']=np.array(jStruct['variancespectrum'])
    specStruct['calibration']=np.array(jStruct['calibration'])
    return specStruct
  spectra=[]
  variances=[]
  skyspecs=[]
  numObjOnFrm=0
  for f in specElem['frame']:
    print "loading ",f," for spectral extraction of object ",name,"."
    path,fn,ext = SpecReduceUtils.splitPathExt(f)
    frm = SpecReduceUtils.Frame(f,type='data',savePath=specdir,frameSetup=framesetup,loadMode='badpixDebias',bias=bias)
    string="_Debiased"
    #print "DEBUG:  having files writing to locations that don't make much sense.  directory: ",specdir,"    string: ",string
    if (flat is not None):
      flattened = SpecReduceUtils.frameDivision(frm,flat)
      flattened.Head = copy.deepcopy(frm.Head)
      string = string+"_Flattened"
      flattened.Head['History'] = "Flattened"
      flattened.toFile(os.path.join(specdir,fn+string+".fits"))
      #print "DEBUG:  having files writing to locations that don't make much sense.  directory: ",specdir,"    string: ",string
      frame=flattened
    else:
      frame=frm
    if (USERDefs.debug):
      print "DEBUG: in processSpectrum  decosmic_ctrl:",decosmic_ctrl
    if (decosmic_ctrl):
      string=string+"_Decosmiced"
      decosmiced,cosmicMsk = SpecReduceUtils.removeCosmics(frame)
      decosmiced.toFile(os.path.join(specdir,fn+string+".fits"))
      cosmicMsk.toFile(os.path.join(specdir,fn+"_CosmicMask.fits"))
      #print "DEBUG:  having files writing to locations that don't make much sense.  directory: ",specdir,"    string: ",string
    else:
      decosmiced = flattened
    specs,varis,skyspec = SpecReduceUtils.optimalExtractWrapper(decosmiced,framesetup,mode,name=fn,calibration=cal,savePath=specdir,telluric=telluric,fluxCal=fluxCal)
    #TODO - maybe break calibration to a separate step
    spectra.append(specs)    #this will malfunction when the spectra of the same object in multiple frames are not in the same 
    variances.append(varis)  #  rank in terms of profile brightness
    skyspecs.append(skyspec)
    if len(specs)>numObjOnFrm:
      numObjOnFrm = len(specs)

  speclist=[]
  varlist=[]
  skylist=[]
  specVariPairs=[]
  for i in range(numObjOnFrm):
    speclist.append([])
    varlist.append([])
    skylist.append([])
    specVariPairs.append([])
  for f in spectra:
    for i in range(len(f)):
      speclist[i].append(f[i])
  for f in variances:
    for i in range(len(f)):
      varlist[i].append(f[i])
  for f in skyspecs:
    for i in range(len(f)):
      skylist[i].append(f[i])
  for s in range(len(speclist)):
    #print "DEBUG: building pairs  s:",s
    for f in range(len(speclist[s])):
      #print "DEBUG: building pairs  s:",s,"   f:",f
      specVariPairs[s].append([speclist[s][f],varlist[s][f]])
  specStruct['numSpectra']=len(speclist)
#  for j in range(len(speclist)):
##    if len(starlist) > len(speclist[j]):
##      print "DEBUG: not all frames had same number of extractions. ",len(starlist)," frames yielded ",len(speclist[j])," spectra of rank ",j,"."
#    npSpecs = np.array(speclist[j])
##    print "DEBUG: shape of npSpecs:",npSpecs.shape
#    spec = np.median(npSpecs,0)
#    SpecReduceUtils.savelist(spec,os.path.join(specdir,name+"-spectrum-"+str(j+1)+".dat"))
#    npSkies = np.array(skylist[j])
#    skyspec= np.median(npSkies,0)
#    SpecReduceUtils.savelist(skyspec,os.path.join(specdir,name+"-skySpectrum-"+str(j+1)+".dat"))
#    npVars = np.array(varlist[j])
#    var = np.median(npVars,0)  #TODO - correct way of combining variances
#    SpecReduceUtils.savelist(var,os.path.join(specdir,name+"-varianceSpectrum-"+str(j+1)+".dat"))
#    localCal=cal
#    if (specElem['skyCalCorrect']):
#      offset = SpecReduceUtils.findSkyCalOffset(cal,skyspec,USERDefs.DBSPConstants['goodSkyLines'],framesetup['illuminatedSpan'])
#      localCal=cal+offset
#      print "offsetting calibration by ",offset," because of sky line extracted positions."
#    specStruct['spectrum-'+str(j+1)]=spec
#    specStruct['skyspectrum-'+str(j+1)]=skyspec
#    specStruct['variancespectrum-'+str(j+1)]=var
#    specStruct['calibration-'+str(j+1)]=localCal
#    specPlot("Median spectrum of "+name+"-"+str(j+1),"Ergs/s/cm^2/Angstrom",spec,localCal,specdir,name+'spectrum-'+str(j+1)+".png")
#    specPlot("Variance spectrum of "+name+"-"+str(j+1),"Ergs/s/cm^2/Angstrom",var,localCal,specdir,name+'variance-'+str(j+1)+".png")
#    with open(os.path.join(specdir,name+"-spectrum-"+str(j+1)+".json"), 'w') as f:
#      nthStruct = {'spectrum':spec.tolist(),'skyspectrum':skyspec.tolist(),'variancespectrum':var.tolist(),'calibration':localCal.tolist()}
#      json.dump(nthStruct,f,indent=2,allow_nan=True)
  for j in range(len(specVariPairs)):
    spec,var = SpecReduceUtils.weightedMean(specVariPairs[j])
    skyspec = np.median(np.array(skylist[j]),0)
    if (cal == None):
      cal = np.array(range(spec.size))
    localCal=cal.tolist()
    if (specElem['skyCalCorrect']):
      offset = SpecReduceUtils.findSkyCalOffset(cal,skyspec,USERDefs.DBSPConstants['goodSkyLines'],framesetup['illuminatedSpan'])
      localCal=np.array(cal)+offset
      localCal=localCal.tolist()
      print "offsetting calibration by ",offset," because of sky line extracted positions."
    specStruct['spectrum-'+str(j+1)]=spec
    specStruct['skyspectrum-'+str(j+1)]=skyspec
    specStruct['variancespectrum-'+str(j+1)]=var
    specStruct['calibration-'+str(j+1)]=localCal
    specPlot("Median spectrum of "+name+"-"+str(j+1),"Ergs/s/cm^2/Angstrom",spec,localCal,specdir,name+'spectrum-'+str(j+1)+".png")
    specPlot("Variance spectrum of "+name+"-"+str(j+1),"Ergs/s/cm^2/Angstrom",var,localCal,specdir,name+'variance-'+str(j+1)+".png")
    with open(os.path.join(specdir,name+"-spectrum-"+str(j+1)+".json"), 'w') as f:
      #print "DEBUG try spec: ",spec.tolist()
      #print "DEBUG try skyspec: ",skyspec
      #print "DEBUG try vari: ",var.tolist()
      #print "DEBUG try cal: ",localCal
      nthStruct = {'spectrum':spec.tolist(),'skyspectrum':skyspec.tolist(),'variancespectrum':var.tolist(),'calibration':localCal}
      #nthStruct = {'spectrum':spec.tolist(),'skyspectrum':skyspec.tolist(),'variancespectrum':var.tolist(),'calibration':localCal.tolist()}
      json.dump(nthStruct,f,indent=2,allow_nan=True)
    fitstab=os.path.join(specdir,name+"-spectrum-"+str(j+1)+".fits")
    tabhdu=fits.BinTableHDU.from_columns([fits.Column(name="Spectrum",format='E',array=np.array(spec)),
                                          fits.Column(name="Variance",format='E',array=np.array(var)),
                                          fits.Column(name="Wavelengths",format='E',array=np.array(localCal))])
    tabhdu.writeto(fitstab)
  specStruct['spectrum']=specStruct['spectrum-1']
  specStruct['skyspectrum']=specStruct['skyspectrum-1']
  specStruct['variancespectrum']=specStruct['variancespectrum-1']
  specStruct['calibration']=specStruct['calibration-1']
  return specStruct


def specPlot(title,ylabel,spec,cal,saveDir,saveName):
  #print "DEBUG: plot sizes: spec:",len(spec)," cal: ",len(cal)
  if (cal is None):
    cal = range(len(spec))
  plt.title(title)
  plt.ylabel(ylabel)
  plt.xlabel("wavelength")
  plt.plot(cal,spec,'b-')
  plt.savefig(os.path.join(saveDir,saveName))
  if (USERDefs.display):
    plt.show()
  plt.clf()


def populateTelluric(inputStruct,productsStruct,outputdir,FrmSetup):
  if ('telluric_correct' not in inputStruct):
    return
  for tell in inputStruct['telluric_correct']:
    telldir = setupElementDir(outputdir,tell)
    specElem=None
    if (inputStruct['telluric_correct'][tell]["spectrum"] not in productsStruct):
      spec = inputStruct['telluric_correct'][tell]["spectrum"]
      #print "DEBUG: spec for tell ",tell," is ",spec
      #print "DEBUG: WL_cal for ",spec," is ",inputStruct['spectrum'][spec]['WL_cal']
      specdir = setupElementDir(outputdir,spec)
      bias=getBias(inputStruct,productsStruct,'spectrum',spec)
      flat=getFlat(inputStruct,productsStruct,'spectrum',spec)
      #print "DEBUG calling getCal in telluric"
      cal =getCal( inputStruct,productsStruct,'spectrum',spec)
      #print "DEBUG got a cal of ", cal        
      frmsetup=None
      if ('fpa' in inputStruct['spectrum'][spec]):
        frmsetup=FrmSetup[inputStruct['spectrum'][spec]['fpa']]
      productsStruct[spec]=processSpectrum(spec,inputStruct['spectrum'][spec],specdir,frmsetup,bias,flat,cal,False)
    specElem = productsStruct[inputStruct['telluric_correct'][tell]["spectrum"]]
    productsStruct[tell]=processTelluric(tell,inputStruct['telluric_correct'][tell],telldir,specElem)

def processTelluric(name,tellElem,tellDir,spec=None):
  tellStruct={}
  tellStruct={}
  if "fromfile" in tellElem:
    File = ''
    with open(tellElem['fromfile'],'r') as f:
      File = f.read()
      #json file, pairs specified
    jStruct = json.loads(File)
    tellStruct['telluric_correct']=np.array(jStruct['telluric_correct'])
    tellStruct['telluric_variance']=np.array(jStruct['telluric_variance'])
    tellStruct['calibration']=np.array(jStruct['calibration'])
    return tellStruct
  #print "in processTelluric for ",name," the spec is: ",spec
  #print "spectrum: ",spec['spectrum']
  #print "calibration:",spec['calibration']
  tellStruct['telluric_correct']=SpecReduceUtils.computeTelluric(spec['spectrum'],spec['calibration'],USERDefs.DBSPConstants['telluricRegions'],variance=spec['variancespectrum'])
  tellStruct['telluric_variance']=np.array(spec['variancespectrum'])
  tellStruct['calibration']=np.array(spec['calibration'])
  with open(os.path.join(tellDir,name+"-telluric.json"), 'w') as f:
    outStruct = {'telluric_correct':tellStruct['telluric_correct'].tolist(),'telluric_variance':tellStruct['telluric_variance'].tolist(),'calibration':tellStruct['calibration'].tolist()}
    json.dump(outStruct,f,indent=2,allow_nan=True)
  fitstab=os.path.join(tellDir,name+"-telluric.fits")
  tabhdu=fits.BinTableHDU.from_columns([fits.Column(name="Telluric_Correction",format='E',array=np.array(tellStruct['telluric_correct'])),
                                        fits.Column(name="Telluric_Variance",format='E',array=np.array(tellStruct['telluric_variance'])),
                                        fits.Column(name="Wavelengths",format='E',array=np.array(tellStruct['calibration']))])
  tabhdu.writeto(fitstab)

  return tellStruct


def populateFluxCals(inputStruct,productsStruct,outputdir,FrmSetup):
  if ('flux_calibrate' not in inputStruct):
    return
  for fcal in inputStruct['flux_calibrate']:
    fcaldir = setupElementDir(outputdir,fcal)
    specElem=None
    if (inputStruct['flux_calibrate'][fcal]["spectrum"] not in productsStruct):
      spec = inputStruct['flux_calibrate'][fcal]["spectrum"]
      specdir = setupElementDir(outputdir,spec)
      bias=getBias(inputStruct,productsStruct,'spectrum',spec)
      flat=getFlat(inputStruct,productsStruct,'spectrum',spec)
      cal =getCal( inputStruct,productsStruct,'spectrum',spec)
      frmsetup=None
      if ('fpa' in inputStruct['spectrum'][spec]):
        frmsetup=FrmSetup[inputStruct['spectrum'][spec]['fpa']]
      productsStruct[spec]=processSpectrum(spec,inputStruct['spectrum'][spec],specdir,frmsetup,bias,flat,cal,False)
    specElem = productsStruct[inputStruct['flux_calibrate'][fcal]["spectrum"]]
    productsStruct[fcal]=processFluxCal(fcal,inputStruct['flux_calibrate'][fcal],fcaldir,specElem)


def processFluxCal(name,fluxElem,fluxDir,spec=None):
  #TODO - from and to file
  fluxStruct={}
  if "fromfile" in fluxElem:
    File = ''
    with open(fluxElem['fromfile'],'r') as f:
      File = f.read()
      #json file, pairs specified
    jStruct = json.loads(File)
    fluxStruct['fluxCal']=np.array(jStruct['fluxCal'])
    fluxStruct['fluxVar']=np.array(jStruct['fluxVar'])
    fluxStruct['WLcal']=np.array(jStruct['WLcal'])
    return fluxStruct
  ref = SpecReduceUtils.loadFluxStandard(fluxElem['reference'])
  fluxStruct['fluxCal'] = SpecReduceUtils.computeFluxCal(spec['spectrum'],spec['calibration'],ref)
  fluxStruct['fluxVar'] = spec['variancespectrum']
  fluxStruct['WLcal']=spec['calibration']
  with open(os.path.join(fluxDir,name+"-fluxCal.json"), 'w') as f:
    outStruct = {'fluxCal':fluxStruct['fluxCal'].tolist(),'fluxVar':fluxStruct['fluxVar'].tolist(),'WLcal':fluxStruct['WLcal']}
    json.dump(outStruct,f,indent=2,allow_nan=True)
  fitstab=os.path.join(fluxDir,name+"-fluxCal.fits")
  tabhdu=fits.BinTableHDU.from_columns([fits.Column(name="Flux_Calibration",format='E',array=np.array(fluxStruct['fluxCal'])),
                                        fits.Column(name="Flux_Variance",format='E',array=np.array(fluxStruct['fluxVar'])),
                                        fits.Column(name="Wavelengths",format='E',array=np.array(fluxStruct['WLcal']))])
  tabhdu.writeto(fitstab)
  return fluxStruct
  



def mergeSpectra(inputStruct,productsStruct,outputdir,FrmSetup):
  if ('mergeSpectra' not in inputStruct):
    return
  for merg in inputStruct['mergeSpectra']:
    mergdir = setupElementDir(outputdir,merg)
    productsStruct[merg]=processMerge(merg,mergdir,inputStruct['mergeSpectra'][merg],productsStruct)

def processMerge(name,mergdir,mergElem,productsStruct):
  mergStruct={}
  if (USERDefs.debug):
    print "DEBUG: merging spectra: control structure:",mergElem
  #DEBUG
  for k in productsStruct:
    if (USERDefs.debug):
      print "DEBUG: key:",k
  #/DEBUG
  spec1=productsStruct[mergElem['blue']]
  spec2=productsStruct[mergElem['red']]
  mergStruct['spectrum'],mergStruct['calibration'] = SpecReduceUtils.mergeSpecs(spec1['spectrum'],spec1['calibration'],spec2['spectrum'],spec2['calibration'])
  mergStruct['variancespectrum'],foo = SpecReduceUtils.mergeSpecs(spec1['variancespectrum'],spec1['calibration'],spec2['variancespectrum'],spec2['calibration'])
  specPlot("spectrum of "+name,"Ergs/s/cm^2/Angstrom",mergStruct['spectrum'],mergStruct['calibration'],mergdir,name+"spectrum.png")
  specPlot("variance of "+name,"variance",mergStruct['variancespectrum'],mergStruct['calibration'],mergdir,name+"variance.png")
  with open(os.path.join(mergdir,name+"-spectrum.json"), 'w') as f:
    nthStruct = {'spectrum':mergStruct['spectrum'].tolist(),'variancespectrum':mergStruct['variancespectrum'].tolist(),'calibration':mergStruct['calibration'].tolist()}
    json.dump(nthStruct,f,indent=2,allow_nan=True)
  fitstab=os.path.join(mergdir,name+"-spectrum.fits")
  tabhdu=fits.BinTableHDU.from_columns([fits.Column(name="Spectrum",format='E',array=np.array(mergStruct['spectrum'])),
                                        fits.Column(name="Variance",format='E',array=np.array(mergStruct['variancespectrum'])),
                                        fits.Column(name="Wavelengths",format='E',array=np.array(mergStruct['calibration']))])
  tabhdu.writeto(fitstab)
  return mergStruct






if __name__ == "__main__":
  main(sys.argv[1:])


